# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius F. Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip


import logging
import os
import sys
import time
import getpass
import traceback
from collections import OrderedDict

import numpy as np
import scipy.misc as misc
import matplotlib.pyplot as plt

import theano

from ..neuromancer.model import modelload, rebuild_model
from ..neuromancer.loss import SquaredLoss
from .. import utils
from ..config import config
from ..utils.plotting import plot_trainingtarget, my_quiver
from ..utils.locking import FileLock
from ..data import transformations
from . import trainutils
from .parallelisation import BackgroundProc
from .trainutils import HistoryTracker, Schedule

floatX = theano.config.floatX

logger = logging.getLogger('elektronn2log')
inspection_logger = logging.getLogger('elektronn2log-inspection')
user_name = getpass.getuser()
__all__ = ['Trainer', 'TracingTrainer', 'TracingTrainerRNN']


NLL_TEXT = "The NN diverged to `nan` Loss!!!\n "\
           "You have the chance to inspect the last used examples and the "\
           "internal state of the pipeline in the command line. The last "\
           "presented training input data is `batch[0]` and the "\
           "corresponding target `batch[1]`"\

class Trainer(object):
    def __init__(self, exp_config):
        self.exp_config = exp_config
        self.schedules = OrderedDict()
        self.model      = self._create_model()
        self.data       = self._load_data()
        self.batch_size = self._infer_batch_size()
        self.tracker    = HistoryTracker()
        self.tracker.register_debug_output_names(self.model.debug_output_names)

        self.preview_data = self._load_preview_data()
        self.saved_raw_preview = False

        self.get_batch_kwargs = self.exp_config.data_batch_args
        self.get_batch_kwargs['batch_size'] = self.batch_size
        self.get_batch_kwargs['source'] = 'train'

        if self.exp_config.monitor_batch_size%self.batch_size!=0:
            bs = int(np.ceil(float(self.exp_config.monitor_batch_size)/self.batch_size))
            self.exp_config.monitor_batch_size = bs


        os.chdir(self.exp_config.save_path) # The trainer works directly in the save dir



        #self.debug_store = []


    def _create_model(self):
        if self.exp_config.create_model:
            if self.exp_config.model_load_args:
                mdl =  self.exp_config.create_model(
                    self.exp_config.model_load_args)
            else:
                mdl =  self.exp_config.create_model()
        else:
            mdl =  modelload(self.exp_config.model_load_path,
                             **self.exp_config.model_load_args)

        mdl.set_opt_meta_params(self.exp_config.optimiser,
                                       self.exp_config.optimiser_params)

        for var, params in self.exp_config.schedules.items():
            if params:
                schedule = Schedule(**params)
                try:
                    schedule.bind_variable(obj=mdl, prop_name=var)
                except:
                    logger.debug("%s not found in model, trying config now" %(var))
                    schedule.bind_variable(obj=self.exp_config, prop_name=var)

                self.schedules[var] = schedule
                logger.info(schedule)

        return mdl


    def _load_data(self):
        if isinstance(self.exp_config.data_class, (list, tuple)):
            mod, cls = self.exp_config.data_class
            cls = utils.import_variable_from_file(mod, cls)
        else:
            from .. import data
            cls = getattr(data, self.exp_config.data_class)

        return cls(self.model.input_node, self.model.target_node,
                   **self.exp_config.data_init_kwargs)

    def _load_preview_data(self):
        if self.exp_config.preview_data_path is not None:
            data = utils.h5load(self.exp_config.preview_data_path)
            if not (isinstance(data, list) or  isinstance(data, (tuple, list))):
                data = [data,]
            data = [d.astype(floatX)/d.max() for d in data]
            return data
        else:
            return None

    def _infer_batch_size(self):
        model_bs = self.model.batch_size
        conf_bs  = self.exp_config.batch_size
        if model_bs:
            if conf_bs:
                assert model_bs==conf_bs, "Conflicting batchsizes from " \
                                          "model (%d) and experiment " \
                                          "configuration (%d)" % (model_bs, conf_bs)
            return model_bs
        elif conf_bs:
            return conf_bs


    def run(self):
        exp_config  = self.exp_config
        save_name   = exp_config.save_name
        data        = self.data

        t_passed    = 0
        t_pt        = 2
        t_pi        = 2
        last_save_t = 0
        last_save_t2= 0
        save_time   = config.param_save_h
        save_time2  = config.initial_prev_h

        loss, loss_smooth, train_loss, valid_loss, train_error, valid_error, param_vars = 0, 0, 0, 0, 0, 0, 0
        user_termination = False

        if isinstance(self.model.loss_node.parent[0], SquaredLoss):
            is_regression = True
        else:
            is_regression = False
        pp_err  = 'err' if is_regression else '%'

        # --------------------------------------------------------------------------------------------------------
        if config.background_processes:
            n_proc = max(2, int(config.background_processes))
            bg_worker = BackgroundProc(data.getbatch, n_proc=n_proc, target_kwargs=self.get_batch_kwargs)
        # --------------------------------------------------------------------------------------------------------
        try:
            i = -1
            t0 = time.time()
            while i < exp_config.n_steps:
                print('{:05d}'.format(i), end='\r')
                sys.stdout.flush()
                try:
                    if config.background_processes:
                        batch = bg_worker.get()
                    else:
                        batch = data.getbatch(**self.get_batch_kwargs)

                    if exp_config.class_weights is not None:
                        batch = batch + (exp_config.class_weights,)

                    #self.debug_store.append(batch[1])
                    #-----------------------------------------------------------------------------------------------------
                    loss, t_per_train, debug_outputs = self.model.trainingstep(*batch, optimiser=exp_config.optimiser) # Update step
                    i += 1
                    #-----------------------------------------------------------------------------------------------------
                    t_per_it = time.time() - t0
                    t_passed += t_per_it
                    t0 = time.time()
                    t_pi = 0.8*t_pi + 0.2*t_per_it     # EMA
                    loss_smooth = self.model.loss_smooth

                    # check for divergence
                    if np.any(np.isnan(loss)) or np.any(np.isinf(loss)):
                        logger.warning(NLL_TEXT)
                        raise KeyboardInterrupt
                        #self.self.model.optimisers[exp_config.optimiser].repair()

                    if len(batch) == 1:
                        batch_char = 0
                    else:
                        batch_char = batch[1][:,0].mean() # assuming targets have shape (b,f,...)
                    self.tracker.update_timeline([t_passed, loss, batch_char])
                    if debug_outputs:
                        self.tracker.update_debug_outputs([i, loss,]+debug_outputs)

                    # Save Parameters
                    # if (t_passed-last_save_t)/3600 > config.param_save_h:
                    #     last_save_t = t_passed
                    #     time_string = '-'+str(save_time)+'h'
                    #     self.model.save(os.path.join('Backup', save_name+time_string+'.mdl'))
                    #     save_time += config.param_save_h
                    if i%config.param_save_it==0 and i>0:
                        it_string = '-'+str(i//1000)+'k'
                        self.model.save(os.path.join('Backup', save_name+it_string+'.mdl'))

                    # Create preview prediction images
                    if self.preview_data is not None:
                        if (t_passed-last_save_t2)/3600 > config.prev_save_h \
                            or (t_passed/3600 > config.initial_prev_h and last_save_t2==0): # first time
                            last_save_t2 = t_passed
                            exp_config.preview_kwargs['number'] = save_time2
                            save_time2 += config.prev_save_h
                            try:
                                self.preview_slice(**exp_config.preview_kwargs)
                            except:
                                logger.warning("Preview Predictions failed."
                                               "Are the preview raw data in "
                                               "the correct format?")
                            # reset time because we only count training time
                            # not time spent for previews (making previews
                            # is not a computational payload of the actual
                            # training but just for "fun")
                            t0 = time.time()

                    # Adjust the learning rate and other schedule parameters
                    for schedule in self.schedules.values():
                        if i==schedule.next_update:
                            schedule.update(i)

                    if (i%exp_config.history_freq==0) and exp_config.history_freq!=0:
                        lr = self.model.lr
                        mom = self.model.mom
                        if len(self.model.gradnet_rates):
                            gradnetrate = np.mean(self.model.gradnet_rates)
                        else:
                            gradnetrate = 0

                        ### Training  & Valid Errors ###
                        loss_after = self.model.loss(*batch)
                        loss_gain  = loss_after - loss
                        train_loss, train_error = self.test_model('train')
                        valid_loss, valid_error = self.test_model('valid')
                        if not is_regression:
                            train_error *= 100
                            valid_error *= 100

                        self.tracker.update_history([i, t_passed, train_loss,
                                                     valid_loss, loss_gain,
                                                     train_error, valid_error,
                                                     lr, mom, gradnetrate])

                        ### Plotting / Saving ###
                        self.model.save(save_name+'-LAST.mdl')
                        self.tracker.save(os.path.join('Backup', save_name))
                        if config.plot_on and ((i>=exp_config.history_freq*3) or i>60):
                            self.tracker.plot(save_name)

                        if config.print_status:
                            t = utils.pretty_string_time(t_passed)
                            out = "%05i L_m=%.3f, L=%.2f, tr=%05.2f%s, "%(i,
                                   loss_smooth, loss, train_error, pp_err)
                            out +="vl=%05.2f%s, prev=%04.1f, L_diff=%+.1e, "\
                                   %(valid_error, pp_err, batch_char*100, loss_gain)
                            out +="LR=%.5f, %.2f it/s, %s" %(lr, 1.0/t_pi, t)
                            logger.info(out)

                # User Interface ##############################################
                except (KeyboardInterrupt, ValueError, TypeError) as e:
                    if not isinstance(e, KeyboardInterrupt):
                        traceback.print_exc()
                        print("\nEntering Command line such that Exception can be "
                              "further inspected by user.\n\n")

                    out = "%05i L_m=%.5f, L=%.4f, train=%.5f, valid=%.5f, " %(i,
                           loss_smooth, loss, train_loss, valid_loss)
                    out +="train=%.3f%s, valid=%.3f%s,\n LR=%.6f, MOM=%.6f, "\
                           %(train_error, pp_err, valid_error, pp_err, self.model.lr, self.model.mom)
                    out +="%.1f GPU-it/s, %.1f CPU-it/s, " %(1.0/self.model.time_per_step,
                           1.0/t_pi)

                    t = utils.pretty_string_time(t_passed)
                    logger.info(out+t)

                    # Like a command line, but cannot change singletons
                    var_push = globals()
                    var_push.update(locals())
                    ret = trainutils.user_input(var_push)
                    if ret=='kill':
                        user_termination = True

                    if config.background_processes:
                        bg_worker.reset()

                    plt.close('all')
                    # reset time after user interaction, otherwise time
                    # will appear as pause in plot
                    t0 = time.time()

                    # End UI ##################################################

                # This is in the epoch/UI loop
                if (t_passed > exp_config.max_runtime) or user_termination :
                    logger.info('Timeout or manual Termination')
                    break

            # This is OUTSIDE the training loop i.e.
            # the last block of the function ``run``
            self.model.save(save_name+"-FINAL.mdl")
            if len(self.tracker.timeline) > 10:
                self.tracker.plot(save_name)
            logger.info('End of Training')
            logger.info('#'*60 + '\n' + '#'*60 + '\n')
            # -------------------end of run()----------------------------------

        except:
            sys.excepthook(*sys.exc_info()) # show info on error
        finally:
            if config.background_processes:
                bg_worker.shutdown()

        if self.model.batch_normalisation_active:
            print("Rebuilding model, replacing batch normalisation layers "
                  "with constant values")
            self.model = rebuild_model(self.model, replace_bn='const')


    def test_model(self, data_source):
        """
        Computes Loss and error/accuracy on batch with ``monitor_batch_size``


        Parameters
        ----------

        data_source: string
            'train' or 'valid'

        Returns
        -------
        Loss, error:

        """
        # copy because it is modified in next line!
        kwargs = dict(self.get_batch_kwargs)
        kwargs['source']=data_source
        kwargs['batch_size'] = self.exp_config.monitor_batch_size
        try:
            batch = self.data.getbatch(**kwargs)
        except ValueError:
            logger.warning("Test model, getbatch failed. No validation data?")
            return np.nan, np.nan # 0, 0
        y_aux = []
        if batch[1] is None:
            return 0, 0
        if self.exp_config.class_weights is not None:
            y_aux.append(self.exp_config.class_weights)

        rates = self.model.dropout_rates
        self.model.dropout_rates = ([0.0,]*len(rates))
        batch_axis = self.model.input_node.shape.tag2index('b')
        n = batch[0].shape[batch_axis]
        loss = 0
        error = 0
        for j in range(int(np.ceil(np.float(n)/self.batch_size))):
            slice_obj = [slice(None) for i in range(batch_axis+1)]
            slice_obj[batch_axis] = slice(j*self.batch_size, (j+1)*self.batch_size)
            d = batch[0][slice_obj] # data
            l = batch[1][slice_obj] # target
            if len(batch) > 2:
                aux = []
                for b in batch[2:]:
                    aux.append(b[j*self.batch_size:(j+1)*self.batch_size])

                nl, er, pred =  self.model.predict_ext(d, l, *(aux + y_aux))
            else:
                nl, er, pred =  self.model.predict_ext(d, l, *y_aux)
            nb_samples = d.shape[batch_axis]
            loss += nl*nb_samples
            error += er*nb_samples
        loss /= n
        error /= n
        self.model.dropout_rates = rates # restore old rates
        return loss, error

    def debug_getcnnbatch(self):
        """
        Executes ``getbatch`` but with un-strided labels and always returning
        info. The first batch example is plotted and the whole batch is
        returned for inspection.
        """
        if self.model.ndim>=2:
            kwargs = dict(self.get_batch_kwargs)
            kwargs['force_dense'] = True
            batch = self.data.getbatch(**kwargs)

            data, target = batch[0][0], batch[1][0]
            target[np.isclose(target, -666)] = 0

            if self.model.ndim==2:
                if target.shape[0] >= 3:
                    target = np.transpose(target, (1,2,0))[:,:,:3]
                    target = target[...,[2,1,0]]
                else:
                    target = target[0]
                plot_trainingtarget(data[0], target, 1)
            else:
                t_i = target.shape[1]//2
                if target.shape[0] >= 3:
                    target = np.transpose(target, (1,2,3,0))[0,:,:,:3]
                    target = target[...,[2, 1, 0]]
                else:
                    target = target[0,t_i]
                i = self.data.offsets[0] # z offset
                plot_trainingtarget(data[0,i+t_i], target, 1)

            plt.ion()
            plt.show()
            with FileLock('plotting'):
                plt.savefig('Batch_test_image.png', bbox_inches='tight')
            plt.pause(0.01)
            plt.pause(2.0)
            plt.close('all')
            plt.pause(0.01)
            return batch
        else:
            logger.warning('debug_getcnnbatch() is only available for "img-img"'
                           ' training mode.\nCheck if your prediction node'
                           ' has the right shape.ndim\n(look if'
                           ' model.prediction_node.shape.ndim is >=2.)')


    def predict_and_write(self, pred_node, raw_img, number=0, export_class='all', block_name='', z_thick=5):
        """
        Predict and and save a slice as preview image

        Parameters
        ----------

        raw_img : np.ndarray
          raw data in the format (ch, x, y, z)
        number: int/float
          consecutive number for the save name (i.e. hours, iterations etc.)
        export_class: str or int
          'all' writes images of all classes, otherwise only the
          class with index ``export_class`` (int) is saved.
        block_name: str
          Name/number to distinguish different raw_imges
        """
        block_name = str(block_name)
        pred = pred_node.predict_dense(raw_img) # returns (k, (z,) y, x)
        z_sh = pred.shape[1]
        pred = pred[:,(z_sh-z_thick)//2:(z_sh-z_thick)//2+z_thick,:,:,]
        save_name = self.exp_config.save_name
        with FileLock('plotting'):  # One lock is enough because everything below calls mpl directly
            for z in range(pred.shape[1]):
                if export_class=='all':
                    for c in range(pred.shape[0]):
                        plt.imsave('%s-pred-%s-z%i-c%i-%shrs.png' \
                                   %(save_name, block_name, z, c, number), pred[c,z,:,:], cmap='gray')
                elif export_class in ['malis', 'affinity']:
                    plt.imsave('%s-pred-%s-aff-z%i-%shrs.png' \
                               %(save_name, block_name, z, number),
                               np.transpose(pred[0:6:2,z,:,:],(1,2,0)), cmap='gray')
                else:
                    if isinstance(export_class, (list, tuple)):
                        for c in export_class:
                            plt.imsave('%s-pred-%s-z%i-c%i-%shrs.png' \
                                       %(save_name, block_name, z, c, number), pred[c,z,:,:], cmap='gray')

                    else:
                        c = int(export_class)
                        plt.imsave('%s-pred-%s-z%i-c%i-%shrs.png' \
                                   %(save_name, block_name, z, c, number), pred[c,z,:,:], cmap='gray')

            if not self.saved_raw_preview: # only do once
                if len(pred_node.shape.offsets)==2:
                    z_off = 0
                else:
                    z_off = int(pred_node.shape.offsets[0])

                for z in range(pred.shape[1]):
                    plt.imsave('%s-raw-%s-z%i.png'%(save_name, block_name, z), raw_img[0,z+z_off,:,:], cmap='gray')


    def preview_slice_from_traindata(self, cube_i=0, off=(0,0,0), sh=(10,400,400), number=0, export_class='all'):
        """
        Predict and and save a selected slice from the training data as preview

        Parameters
        ----------

        cube_i: int
          index of source cube in CNNData
        off: 3-tuple of int
          start index of slice to cut from cube (z,y,x)
        sh: 3-tuple of int
          shape of cube to cut (z,y,x)
        number: int
          consecutive number for the save name (i.e. hours, iterations etc.)
        export_class: str or int
          'all' writes images of all classes, otherwise only the class with
          index ``export_class`` (int) is saved.
        """

        if self.model.prediction_node.shape.ndim >= 2:
            pred_node = self.model.prediction_node
        elif "pred_dense" in self.model.nodes:
            pred_node = self.model['pred_dense']
        else:
            raise RuntimeError("Model have spatial prediction node or"
                               " 'pred_dense' node which is spatial")
        if self.model.ndim==3:
            min_z = self.model.prediction_node.input_nodes[0].shape['z']
            if min_z > sh[0]:
                sh = list(sh)
                sh[0] = min_z
        elif self.model.ndim==2:
            pass
        else:
            raise RuntimeError("Model must be 2/3 dimensional for previews")

        raw_img = self.data.train_d[cube_i]
        raw_img = raw_img[:,
                          off[0]:off[0]+sh[0],
                          off[1]:off[1]+sh[1],
                          off[2]:off[2]+sh[2]]

        self.predict_and_write(pred_node, raw_img, number, export_class)
        self.saved_raw_preview = True


    def preview_slice(self, number=0, export_class='all', max_z_pred=5):
        """
        Predict and and save a data from a separately loaded file as preview

        Parameters
        ----------

        number: int/float
          consecutive number for the save name (i.e. hours, iterations etc.)
        export_class: str or int
          'all' writes images of all classes, otherwise only the class with
          index ``export_class`` (int) is saved.
        max_z_pred: int
          approximate maximal number of z-slices to produce (depends on CNN architecture)
        """

        assert self.preview_data is not None, "You must provide preview data in order to call this function"
        for example_no,raw_img in enumerate(self.preview_data):
            if raw_img.ndim==3:
                if raw_img.shape[0]>raw_img.shape[2]:
                    raw_img = np.transpose(raw_img, (2,0,1))
                    logger.warning("preview_slice: transposing preview image, assuming last dim is z because "
                    " this dim is smaller than the first.")

            z_sh = raw_img.shape[0] if raw_img.ndim==3 else raw_img.shape[1]

            if self.model.prediction_node.shape.ndim>=2:
                pred_node =  self.model.prediction_node
            elif "pred_dense" in self.model.nodes:
                pred_node = self.model['pred_dense']
            else:
                raise RuntimeError("Model have spatial prediction node or"
                                   " 'pred_dense' node which is spatial")
            if pred_node.shape.ndim==3:
                strd_z = pred_node.shape.strides[0]
                out_z  = pred_node.shape.spatial_shape[0] * strd_z
                min_z  = pred_node.input_nodes[0].shape.spatial_shape[0] + strd_z - 1 # input shape
                z_thick = min_z if out_z > max_z_pred else min_z + strd_z*int(np.ceil(float(max_z_pred-out_z)/strd_z))
            elif pred_node.shape.ndim==2:
                z_thick = max_z_pred
            else:
                raise RuntimeError("Model must be 2/3 dimensional for previews")


            if z_thick > z_sh:
                raise ValueError("The preview slices are too small in z-direction for this CNN")

            if raw_img.ndim==3:
                raw_img = raw_img[None, (z_sh-z_thick)//2:(z_sh-z_thick)//2+z_thick, :, :]
            elif raw_img.ndim==4:
                raw_img = raw_img[:, (z_sh-z_thick)//2:(z_sh-z_thick)//2+z_thick, :, :]

            self.predict_and_write(pred_node, raw_img, number, export_class, example_no, max_z_pred)

        self.saved_raw_preview = True

###############################################################################

class TracingTrainer(Trainer):
    @staticmethod
    def save_batch(img, lab, k, lab_img=None):
        img = img[0]
        lab = lab[0]
        off = img.shape[1] - lab.shape[1]
        utils.h5save(img, 'img-%i.h5' % k)
        if lab_img is not None:
            utils.h5save(lab_img, 'lab_img-%i.h5' % k)

        # assert off % 2==0
        # off //= 2
        with FileLock('plotting'):
            for i in range(img.shape[1]):
                plt.imsave('batch-%i-z%i.png' % (k, i), img[0, i], cmap='gray')
            # if 0 <= (i - off) < lab.shape[1]:
            #     lab_small = lab[4, i - off]
            #     lab_up = misc.imresize(lab_small,
            #                            np.multiply(lab_small.shape, 8),
            #                            interp='nearest')
            #     plt.imsave('batch-%i-z%i-l.png' % (k, i), lab_up, cmap='gray')


    # def probmap_preview(self, raw_img, number=0, block_name=''):
    #     """
    #     Predict and and save a slice as preview image
    #
    #     Parameters
    #     ----------
    #
    #     raw_img : np.ndarray
    #       raw data in the format (ch, x, y, z)
    #     number: int/float
    #       consecutive number for the save name (i.e. hours, iterations etc.)
    #     block_name: str
    #       Name/number to distinguish different raw_imges
    #     """
    #     block_name = str(block_name)
    #     pred = self.model['pred_dense'].predict_dense(raw_img)  # returns (k, z, y, x)
    #
    #     save_name = self.exp_config.save_name
    #     #names = ['vz', 'vx', 'vy', 'br', 'barr', 'bg', 'syn', 'ves', 'mito']
    #     for z in range(pred.shape[1]):
    #         for c in range(pred.shape[0]):
    #             plt.imsave('%s-pred-%s-z%i-c%i-%shrs.png' \
    #                        % (save_name, block_name, z, c, number),
    #                        pred[c, z, :, :], cmap='gray')
    #
    #     z_off = int(self.model['vec'].shape.offsets[0])
    #     for z in range(pred.shape[1]):
    #         plt.imsave('%s-raw-%s-z%i.png' % (save_name, block_name, z),
    #                    raw_img[0, z + z_off, :, :], cmap='gray')


    def debug_getcnnbatch(self, extended=False):
        """
        Executes ``getbatch`` but with un-strided labels and always returning
        info. The first batch example is plotted and the whole batch is
        returned for inspection.
        """
        kwargs = dict(self.get_batch_kwargs)
        kwargs['force_dense'] = True
        batch = self.data.getbatch(**kwargs)

        data, target = batch[0][0], batch[1][0]
        target[np.isclose(target, -666)] = 0
        if self.model.ndim==2:
            if target.shape[0] >= 3:
                target = np.transpose(target, (1, 2, 0))[:, :, :3]
                target = target[...,[2, 0, 1]]
            else:
                target = target[0]
            plot_trainingtarget(data[0], target, 1)
        else:
            t_i = target.shape[1] // 2
            if target.shape[0] >= 3:
                target = np.transpose(target, (1, 2, 3, 0))[0, :, :, :3]
                target = target[...,[2, 0, 2]]
            else:
                target = target[0, t_i]

            i = self.data.offsets[0]  # z offset
            plot_trainingtarget(data[0, i+t_i], target, 1)


        if self.model.ndim==3 and extended:
            dest = '/tmp/%s-'%user_name
            data, target = batch[0], batch[1]
            target[np.isclose(target, -666)] = 0
            i = self.data.offsets[0]  # z offset
            with FileLock('plotting'):
                for j in range(data.shape[2]):
                    plt.imsave('/tmp/img-%i.png' % j, data[0, 0, j], cmap='gray')
                    if j - i >= 0 and j - i < target.shape[2]:
                        plt.imsave(dest+'img-%i-br.png'%j, target[0, 4, j - i],cmap='gray')
                        plt.imsave(dest+'img-%i-z.png'%j, target[0,0,j-i], cmap='gray')
                        plt.imsave(dest+'img-%i-y.png'%j, target[0,1,j-i], cmap='gray')
                        plt.imsave(dest+'img-%i-x.png'%j, target[0,2,j-i], cmap='gray')
                        plt.imsave(dest+'img-%i-barr.png'%j, target[0,3,j-i], cmap='gray')

                        plt.imsave(dest+'img-%i-syn.png'%j, target[0,6,j-i], cmap='gray')
                        plt.imsave(dest+'img-%i-ves.png'%j, target[0,7,j-i], cmap='gray')
                        plt.imsave(dest+'img-%i-mito.png'%j, target[0,8,j-i], cmap='gray')

                        quiver = my_quiver(target[0,2,j-i], target[0,1,j-i],
                                  img=target[0, 4, j - i], c=target[0,0,j-i])

                        quiver.savefig(dest+'vec-%i.png'%j, bbox_inches='tight')

        plt.ion()
        plt.show()
        with FileLock('plotting'):
            plt.savefig('Batch_test_image.png', bbox_inches='tight')
        plt.pause(0.01)
        plt.pause(2.0)
        plt.close('all')
        plt.pause(0.01)

        return batch

    def run(self):
        exp_config = self.exp_config
        save_name = exp_config.save_name
        data = self.data
        self.tracker.register_debug_output_names(self.model.debug_output_names[1:]) #remove first because it is prediction

        t_passed = 0
        t_pt = 2
        t_pi = 2
        last_save_t = 0
        last_save_t2 = 0
        save_time = config.param_save_h
        save_time2 = config.initial_prev_h

        loss, loss_smooth, train_loss, valid_loss, train_error, valid_error, param_vars = 0, 0, 0, 0, 0, 0, 0
        user_termination = False

        if isinstance(self.model.loss_node.parent[0], SquaredLoss):
            is_regression = True
        else:
            is_regression = False
        pp_err = 'err' if is_regression else '%'

        # --------------------------------------------------------------------------------------------------------
        if config.background_processes:
            n_proc = max(2, int(config.background_processes))
            bg_worker = BackgroundProc(data.getbatch, n_proc=n_proc,
                                       target_kwargs=self.get_batch_kwargs)
        # --------------------------------------------------------------------------------------------------------
        try:
            lost_track = True
            tracing_length = 1

            i = -1
            t0 = time.time()
            while i < exp_config.n_steps:
                # update max every loop to make it modifyable during training
                max_tracing = exp_config.sequence_training if exp_config.sequence_training>1 else 50
                try:
                    if config.inspection:
                        inspection_logger.info("#BATCH %i" % (i + 1))

                    # check if we lost the skeleton track
                    if exp_config.sequence_training:
                        try:
                            lost_track= batch[2].lost_track
                        except UnboundLocalError:
                            pass
                        if tracing_length >= max_tracing:
                            lost_track = True

                    # if we are still on track get a new slice from this skeleton
                    if not lost_track and exp_config.sequence_training:
                        tracing_length  += 1
                        position_l, direction_il = batch[3].cnn_pred2lab_position(
                            prediction_c)
                        try:
                            tmp = data.get_newslice(position_l, direction_il,
                                                           **self.get_batch_kwargs)
                            img, vec, trafo = tmp[:3]
                            if len(tmp)==4:
                                batch = (img, vec, batch[2], trafo, tmp[3])
                            else:
                                batch = (img, vec, batch[2], trafo)
                        # if get_newslice fails, do same as when track is lost
                        except transformations.WarpingOOBError:
                            lost_track = True

                    if lost_track and exp_config.sequence_training:
                        print("Traced for %i iterations" %(tracing_length,))
                        tracing_length = 0
                        try:
                            skel = batch[2]
                            skel.debug_traces.append(np.array(skel.debug_traces_current))
                            skel.debug_traces_current = []
                            skel.debug_grads.append(np.array(skel.debug_grads_current))
                            skel.debug_grads_current = []
                        except UnboundLocalError:
                            pass

                    # non-sequence training or getting a new skeleton
                    if lost_track or (not exp_config.sequence_training):
                        if config.background_processes:
                            batch = bg_worker.get()
                        else:
                            batch = data.getbatch(**self.get_batch_kwargs)

                        if exp_config.sequence_training:
                            #print("Next skel: %i" % (batch[2],))
                            pass

                        batch = list(batch)
                        batch[2] = data.train_s[batch[2]]
                        batch[3] = transformations.trafo_from_array(batch[3][0])

                    if config.inspection:
                        lab_img = batch[4]
                        batch = batch[:4]
                        if (i+1) % 50==0:
                            self.save_batch(batch[0], batch[1], (i + 1), lab_img)

                    # -----------------------------------------------------------------------------------------------------
                    loss, t_per_train, debug_outputs = self.model.trainingstep(*batch, optimiser=exp_config.optimiser)  # Update step
                    prediction_c = debug_outputs[0][0]
                    debug_outputs = debug_outputs[1:]
                    i += 1
                    # -----------------------------------------------------------------------------------------------------
                    t_per_it = time.time() - t0
                    t_passed += t_per_it
                    t0 = time.time()
                    t_pi = 0.8 * t_pi + 0.2 * t_per_it  # EMA
                    loss_smooth = self.model.loss_smooth


                    # check for divergence
                    if np.any(np.isnan(loss)) or np.any(np.isinf(loss)):
                        logger.warning(NLL_TEXT)
                        raise KeyboardInterrupt
                        # self.model.optimisers[exp_config.optimiser].repair()

                    self.tracker.update_timeline([t_passed, loss, debug_outputs[0]/10])
                    if debug_outputs:
                        self.tracker.update_debug_outputs([i, loss, ] + debug_outputs)

                    # Save Parameters
                    # if (t_passed-last_save_t)/3600 > config.param_save_h:
                    #     last_save_t = t_passed
                    #     time_string = '-'+str(save_time)+'h'
                    #     self.model.save(os.path.join('Backup', save_name+time_string+'.mdl'))
                    #     save_time += config.param_save_h
                    if i%config.param_save_it==0 and i>0:
                        it_string = '-'+str(i//1000)+'k'
                        self.model.save(os.path.join('Backup', save_name+it_string+'.mdl'))

                    # Create preview prediction images
                    if self.preview_data is not None:
                        if (t_passed - last_save_t2) / 3600 > config.prev_save_h \
                           or (t_passed / 3600 > config.initial_prev_h
                           and last_save_t2==0):  # first time
                            last_save_t2 = t_passed
                            exp_config.preview_kwargs['number'] = save_time2
                            save_time2 += config.prev_save_h
                            try:
                                self.preview_slice(**exp_config.preview_kwargs)
                            except:
                                logger.warning("Preview Predictions failed."
                                               "Are the preview raw data in "
                                               "the correct format?")
                            # reset time because we only count training time
                            # not time spent for previews (making previews
                            # is not a computational payload of the actual
                            # training but just for "fun")
                            t0 = time.time()

                    # Adjust the learning rate and other schedule parameters
                    for schedule in self.schedules.values():
                        if i==schedule.next_update:
                            schedule.update(i)

                    if (i % exp_config.history_freq==0) \
                        and exp_config.history_freq!=0:
                        lr = self.model.lr
                        mom = self.model.mom
                        if len(self.model.gradnet_rates):
                            gradnetrate = np.mean(self.model.gradnet_rates)
                        else:
                            gradnetrate = 0

                        ### Training  & Valid Errors ###
                        loss_after = self.model.loss(*batch)
                        loss_gain = loss_after - loss
                        train_loss, train_error = self.test_model('train')
                        valid_loss, valid_error = self.test_model('valid')
                        self.tracker.update_history([i, t_passed, train_loss,
                                                     valid_loss, loss_gain,
                                                     train_error, valid_error,
                                                     lr, mom, gradnetrate])

                        ### Plotting / Saving ###
                        self.model.save(save_name + '-LAST.mdl')
                        self.tracker.save(os.path.join('Backup', save_name))
                        if config.plot_on and ((i>=exp_config.history_freq*3) or i>60):
                            self.tracker.plot(save_name)

                        if config.print_status:
                            t = utils.pretty_string_time(t_passed)
                            out = "%05i L_m=%.3f, L=%.2f, tr=%05.2f%s, " % (i,
                                                                            loss_smooth,
                                                                            loss,
                                                                            train_error,
                                                                            pp_err)
                            out += "vl=%05.2f%s, prev=%04.1f, L_diff=%+.1e, " \
                                   % (valid_error, pp_err, debug_outputs[0] * 100,
                                      loss_gain)
                            out += "LR=%.5f, %.2f it/s, %s" % (
                            lr, 1.0 / t_pi, t)
                            logger.info(out)

                # User Interface ##############################################
                except (KeyboardInterrupt, ValueError, TypeError) as e:
                    if not isinstance(e, KeyboardInterrupt):
                        traceback.print_exc()
                        print(
                            "\nEntering Command line such that Exception can be "
                            "further inspected by user.\n\n")

                    out = "%05i L_m=%.5f, L=%.4f, train=%.5f, valid=%.5f, " % (
                    i,
                    loss_smooth, loss, train_loss, valid_loss)
                    out += "train=%.3f%s, valid=%.3f%s,\n LR=%.6f, MOM=%.6f, " \
                           % (
                           train_error, pp_err, valid_error, pp_err, self.model.lr,
                           self.model.mom)
                    out += "%.1f GPU-it/s, %.1f CPU-it/s, " % (
                    1.0 / self.model.time_per_step,
                    1.0 / t_pi)

                    t = utils.pretty_string_time(t_passed)
                    logger.info(out + t)

                    # Like a command line, but cannot change singletons
                    var_push = globals()
                    var_push.update(locals())
                    ret = trainutils.user_input(var_push)
                    if ret=='kill':
                        user_termination = True

                    if config.background_processes:
                        bg_worker.reset()

                    plt.close('all')
                    # reset time after user interaction, otherwise time
                    # will appear as pause in plot
                    t0 = time.time()

                    # End UI ##################################################

                # This is in the epoch/UI loop
                if (t_passed > exp_config.max_runtime) or user_termination:
                    logger.info('Timeout or manual Termination')
                    break

            # This is OUTSIDE the training loop i.e.
            # the last block of the function ``run``
            self.model.save(save_name + "-FINAL.mdl")
            if len(self.tracker.timeline) > 10:
                self.tracker.plot(save_name)
            logger.info('End of Training')
            logger.info('#' * 60 + '\n' + '#' * 60 + '\n')
            # -------------------end of run()----------------------------------

        except:
            sys.excepthook(*sys.exc_info())  # show info on error
        finally:
            if config.background_processes:
                bg_worker.shutdown()

        if self.model.batch_normalisation_active:
            self.model = rebuild_model(self.model, replace_bn='const')

    def test_model(self, data_source):
        """
        Computes Loss and error/accuracy on batch with ``monitor_batch_size``


        Parameters
        ----------

        data_source: string
            'train' or 'valid'

        Returns
        -------
        Loss, error:

        """
        assert self.batch_size==1
        # copy because it is modified in next line!
        kwargs = dict(self.get_batch_kwargs)
        kwargs['source'] = data_source
        kwargs['batch_size'] = self.exp_config.monitor_batch_size
        try:
            batch = self.data.getbatch(**kwargs)
        except ValueError:
            logger.warning("Test model, getbatch failed. No validation data?")
            return np.nan, np.nan  # 0, 0

        batch = list(batch)
        batch[2] = list(batch[2])
        batch[3] = list(batch[3])
        for i in range(self.exp_config.monitor_batch_size):
            if data_source=='train':
                batch[2][i] = self.data.train_s[batch[2][i]]
            elif data_source=='valid':
                batch[2][i] = self.data.valid_s[batch[2][i]]

            batch[3][i] = transformations.trafo_from_array(batch[3][i])

        if config.inspection:
            batch = batch[:4]

        rates = self.model.dropout_rates
        self.model.dropout_rates = ([0.0, ] * len(rates))
        n = len(batch[0])
        loss = 0
        error = 0
        for j in range(n):
            d = batch[0][j:j+1]  # data
            l = batch[1][j:j+1]  # target
            if len(batch) > 2:
                aux = []
                for b in batch[2:]:
                    aux.append(b[j])

                nl, er, pred = self.model.predict_ext(d, l, *aux)
                skel = batch[2][j] # predict_ext calls get_loss_and_gradient
                # which adds current position and grad, but for testing model
                # we do not want these tracked --> remove again
                skel.debug_traces_current.pop()
                skel.debug_grads_current.pop()
            else:
                nl, er, pred = self.model.predict_ext(d, l)

            loss += nl * len(d)
            error += er * len(d)

        loss /= n
        error /= n
        self.model.dropout_rates = rates  # restore old rates
        return loss, error


###############################################################################

class TracingTrainerRNN(TracingTrainer):
    def run(self):
        exp_config = self.exp_config
        save_name = exp_config.save_name
        data = self.data
        if 'scan_out_radius_t'==self.model.debug_output_names[-1]:
            assert 'scan_out_radius'==self.model.debug_output_names[-2]
            print("Found regression targets for scan")
            self.tracker.register_debug_output_names(self.model.debug_output_names[:-2])

        mem_hid = np.zeros(self.model['mem_hid'].shape, dtype=floatX)

        t_passed = 0
        t_pt = 2
        t_pi = 2
        last_save_t = 0
        last_save_t2 = 0
        save_time = config.param_save_h
        save_time2 = config.initial_prev_h

        loss, loss_smooth, train_loss, valid_loss, train_error, valid_error, param_vars = 0, 0, 0, 0, 0, 0, 0
        user_termination = False

        is_regression = True
        pp_err = 'err'

        # --------------------------------------------------------------------------------------------------------
        try:
            lost_track = True
            tracing_length = 0

            i = -1
            t0 = time.time()
            skel_example = None
            while i < exp_config.n_steps:
                # update max every loop to make it modifyable during training
                max_tracing = exp_config.sequence_training
                try:
                    if skel_example is None:
                        skel_example, skel_index = data.getskel('train')
                        if len(self.model.loss_node.input_nodes)==2:
                            batch = (skel_example, mem_hid)
                        else:
                            batch = (skel_example,)
                        skel_example.start_new_training = True
                        tracing_length = 0
                        if config.inspection:
                            inspection_logger.info("NEW SKEL %i" %skel_index)

                    # -----------------------------------------------------------------------------------------------------
                    try:
                        loss, t_per_train, debug_outputs = self.model.trainingstep(*batch, optimiser=exp_config.optimiser)  # Update step
                        tracing_length += 1
                        i += 1
                        if config.inspection:
                            inspection_logger.info("-"*20)
                    except transformations.WarpingOOBError:
                        if config.inspection:
                            inspection_logger.info("OOB, Traced for %i iterations" % (tracing_length,))
                        skel_example = None
                        continue
                    # -----------------------------------------------------------------------------------------------------
                    t_per_it = time.time() - t0
                    t_passed += t_per_it
                    t0 = time.time()
                    t_pi = 0.8 * t_pi + 0.2 * t_per_it  # EMA
                    loss_smooth = self.model.loss_smooth

                    if skel_example.lost_track or tracing_length >= max_tracing:
                        if config.inspection:
                            inspection_logger.info("Traced for %i iterations" % (tracing_length,))
                        skel_example = None

                    # check for divergence
                    if np.any(np.isnan(loss)) or np.any(np.isinf(loss)):
                        logger.warning(NLL_TEXT)
                        raise KeyboardInterrupt
                        # self.model.optimisers[exp_config.optimiser].repair()

                    self.tracker.update_timeline([t_passed, loss, 0])
                    if debug_outputs:
                        if 'scan_out_radius_t'==self.model.debug_output_names[-1]:
                            r = debug_outputs[-2]
                            r_t = debug_outputs[-1]
                            debug_outputs = debug_outputs[:-2]
                            self.tracker.update_regression(r.ravel(), r_t.ravel())

                        debug_outputs_ = [np.mean(x) for x in debug_outputs]
                        self.tracker.update_debug_outputs([i, loss, ] + debug_outputs_)

                    # Save Parameters
                    # if (t_passed-last_save_t)/3600 > config.param_save_h:
                    #     last_save_t = t_passed
                    #     time_string = '-'+str(save_time)+'h'
                    #     self.model.save(os.path.join('Backup', save_name+time_string+'.mdl'))
                    #     save_time += config.param_save_h
                    if i%config.param_save_it==0 and i>0:
                        it_string = '-'+str(i//1000)+'k'
                        self.model.save(os.path.join('Backup', save_name+it_string+'.mdl'))

                    # Create preview prediction images
                    if self.preview_data is not None:
                        if (t_passed - last_save_t2) / 3600 > config.prev_save_h \
                           or (t_passed / 3600 > config.initial_prev_h
                           and last_save_t2==0):  # first time
                            last_save_t2 = t_passed
                            exp_config.preview_kwargs['number'] = save_time2
                            save_time2 += config.prev_save_h
                            try:
                                self.preview_slice(**exp_config.preview_kwargs)
                            except:
                                logger.warning("Preview Predictions failed."
                                               "Are the preview raw data in "
                                               "the correct format?")
                            # reset time because we only count training time
                            # not time spent for previews (making previews
                            # is not a computational payload of the actual
                            # training but just for "fun")
                            t0 = time.time()

                    # Adjust the learning rate and other schedule parameters
                    for schedule in self.schedules.values():
                        if i==schedule.next_update:
                            schedule.update(i)

                    if (i % exp_config.history_freq==0) \
                        and exp_config.history_freq!=0:
                        lr = self.model.lr
                        mom = self.model.mom
                        if len(self.model.gradnet_rates):
                            gradnetrate = np.mean(self.model.gradnet_rates)
                        else:
                            gradnetrate = 0

                        ### Training  & Valid Errors ###
                        loss_after = loss # too expensive to really compute for RNN
                        loss_gain = loss_after - loss
                        train_loss, train_error = self.test_model('train')
                        valid_loss, valid_error = self.test_model('valid')
                        self.tracker.update_history([i, t_passed, train_loss,
                                                     valid_loss, loss_gain,
                                                     train_error, valid_error,
                                                     lr, mom, gradnetrate])

                        ### Plotting / Saving ###
                        self.model.save(save_name + '-LAST.mdl')
                        self.tracker.save(os.path.join('Backup', save_name))
                        if config.plot_on and ((i>=exp_config.history_freq*3) or i>60):
                            self.tracker.plot(save_name)

                        if config.print_status:
                            t = utils.pretty_string_time(t_passed)
                            out = "%05i L_m=%.3f, L=%.2f, tr=%05.2f%s, "% (i,
                            loss_smooth, loss, train_error, pp_err)
                            out += "vl=%05.2f%s, prev=%04.1f, L_diff=%+.1e, " \
                                   % (valid_error, pp_err, 0 * 100,
                                      loss_gain)
                            out += "LR=%.5f, %.2f it/s, %s" % (
                            lr, 1.0 / t_pi, t)
                            logger.info(out)

                # User Interface ##############################################
                except (KeyboardInterrupt, ValueError, TypeError) as e:
                    if not isinstance(e, KeyboardInterrupt):
                        traceback.print_exc()
                        print(
                            "\nEntering Command line such that Exception can be "
                            "further inspected by user.\n\n")

                    out = "%05i L_m=%.5f, L=%.4f, train=%.5f, valid=%.5f, " % (
                    i,
                    loss_smooth, loss, train_loss, valid_loss)
                    out += "train=%.3f%s, valid=%.3f%s,\n LR=%.6f, MOM=%.6f, " \
                           % (
                           train_error, pp_err, valid_error, pp_err, self.model.lr,
                           self.model.mom)
                    out += "%.1f GPU-it/s, %.1f CPU-it/s, " % (
                    1.0 / self.model.time_per_step,
                    1.0 / t_pi)

                    t = utils.pretty_string_time(t_passed)
                    logger.info(out + t)

                    # Like a command line, but cannot change singletons
                    var_push = globals()
                    var_push.update(locals())
                    ret = trainutils.user_input(var_push)
                    if ret=='kill':
                        user_termination = True

                    plt.close('all')
                    # reset time after user interaction, otherwise time
                    # will appear as pause in plot
                    t0 = time.time()

                    # End UI ##################################################

                # This is in the epoch/UI loop
                if (t_passed > exp_config.max_runtime) or user_termination:
                    logger.info('Timeout or manual Termination')
                    break

            # This is OUTSIDE the training loop i.e.
            # the last block of the function ``run``
            self.model.save(save_name + "-FINAL.mdl")
            if len(self.tracker.timeline) > 10:
                self.tracker.plot(save_name)
            logger.info('End of Training')
            logger.info('#' * 60 + '\n' + '#' * 60 + '\n')
            # -------------------end of run()----------------------------------

        except:
            sys.excepthook(*sys.exc_info())  # show info on error
        finally:
            pass

        # if self.model.batch_normalisation_active:
        #     self.model = rebuild_model(self.model, replace_bn='const')



    def test_model(self, data_source):
        #return 1.0, 0.5
        """
        Computes Loss and error/accuracy on batch with ``monitor_batch_size``


        Parameters
        ----------

        data_source: string
            'train' or 'valid'

        Returns
        -------
        Loss, error:

        """
        try:
            skel_example, skel_index = self.data.getskel(data_source)
        except ValueError:
            logger.warning("Test model, getbatch failed. No validation data?")
            return np.nan, np.nan  # 0, 0

        rates = self.model.dropout_rates
        self.model.dropout_rates = ([0.0, ] * len(rates))

        n = self.exp_config.monitor_batch_size
        loss = 0
        j = 0
        while j < n:
            #print(j, end="")
            try:
                skel_example, skel_index = self.data.getskel(data_source)
                skel_example.start_new_training = True
                nl = self.model.loss(skel_example)
            except transformations.WarpingOOBError:
                continue
            j += 1
            loss += nl

        loss /= n
        self.model.dropout_rates = rates  # restore old rates
        return loss, loss



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, \
    super, zip


import multiprocessing as mp
import ctypes
import logging
import time
from collections import deque
import numpy as np


__all__ = ['BackgroundProc', 'SharedQ', 'TimeoutError']

logger = logging.getLogger('elektronn2log')


class TimeoutError(RuntimeError):
    def __init__(self, *args, **kwargs):
        super(TimeoutError, self).__init__(*args, **kwargs)


class SharedMem(object):
    """Utilities to share np.arrays between processes"""
    _ctypes_to_numpy = {ctypes.c_int8: np.dtype(np.int8),
                        ctypes.c_int16: np.dtype(np.int16),
                        ctypes.c_uint16: np.dtype(np.uint16),
                        ctypes.c_int32: np.dtype(np.int32),
                        ctypes.c_uint32: np.dtype(np.uint32),
                        ctypes.c_int64: np.dtype(np.int64),
                        ctypes.c_uint64: np.dtype(np.uint64),

                        ctypes.c_byte: np.dtype(np.int8),
                        ctypes.c_ubyte: np.dtype(np.uint8),
                        ctypes.c_short: np.dtype(np.int16),
                        ctypes.c_ushort: np.dtype(np.uint16),
                        ctypes.c_int: np.dtype(np.int32),
                        ctypes.c_uint: np.dtype(np.uint32),
                        ctypes.c_long: np.dtype(np.int32),
                        ctypes.c_ulong: np.dtype(np.uint32),
                        ctypes.c_longlong: np.dtype(np.int64),
                        ctypes.c_ulonglong: np.dtype(np.int64),

                        ctypes.c_float: np.dtype(np.float32),
                        ctypes.c_double: np.dtype(np.float64)}

    _numpy_to_ctypes = dict(
        list(zip(_ctypes_to_numpy.values(), _ctypes_to_numpy.keys())))

    @staticmethod
    def shm2ndarray(mp_array, shape=None):
        """
        Parameters
        ----------

        mp_array: a mp.Array
        shape: (optional) the returned np.ndarray
                is reshaped to this shape, flat otherwise

        Returns
        -------

        array: np.ndarray
         That can be normally used but changes are reflected in shared mem

        Note: the returned array is still pointing to the sharedmem,
              data might be changed by another process!
        """
        dtype = SharedMem._ctypes_to_numpy[mp_array._type_]
        result = np.frombuffer(mp_array, dtype)

        if shape is not None:
            result = result.reshape(shape)

        return np.asarray(result)

    @staticmethod
    def ndarray2shm(np_array, lock=False):
        """
        Parameters
        ----------

        np_array: np.ndarray
          array of arbitrary shape
        lock: Bool
          Whether to create a multiprocessing.Lock

        Returns
        -------

        handle: mp.Array:
          flat with data from ndarray copied to it
        """
        array1d = np_array.ravel(order='A')

        try:
            c_type = SharedMem._numpy_to_ctypes[array1d.dtype]
        except KeyError:
            c_type = SharedMem._numpy_to_ctypes[np.dtype(array1d.dtype)]

        result = mp.Array(c_type, array1d.size, lock=lock)
        SharedMem.shm2ndarray(result)[:] = array1d
        return result

    def puthandle(self, dtype, shape, data=None, lock=False):
        """
        Creates new shared memory and puts it on the queue.
        Other sub-processes can write to it.

        Parameters
        ----------
        dtype: np.dtype
          Type of data to store in array
        shape: tuple
          Properties of shared mem to be created
        data: np.ndarray
         (optional) values to fill shared array with
        lock: Bool
          Whether to create a multiprocessing.Lock on the shared variable

        Returns
        -------
        sharedmem handle: mp.array
        """
        t0 = time.clock()
        size = int(np.prod(shape))
        try:
            c_type = SharedMem._numpy_to_ctypes[dtype]
        except KeyError:
            c_type = SharedMem._numpy_to_ctypes[np.dtype(dtype)]

        shm = mp.Array(c_type, size, lock=lock)
        t1 = time.clock()

        if data is not None:
            SharedMem.shm2ndarray(shm)[:] = data.astype(dtype).ravel(order='A')
        t2 = time.clock()

        if self.profile:
            t_alloc = t1 - t0
            t_write = t2 - t1
            self.logger.info('SharedMemAlloc %g ms, WriteInitialData %g ms' % (
                t_alloc * 1000, t_write * 1000))

        return shm


class Proc(mp.Process):
    """
    A *reusable* and *configurable* background process, that does the
    same job every time ``events['new']`` is set and signals that is has
    finished one iteration by setting ``events['ready']``
    """

    def __init__(self, mp_arrays, shapes, events, target, target_args,
                 target_kwargs, profile):
        super(Proc, self).__init__()
        self.events = events
        self.target = target
        self.target_args = target_args
        self.target_kwargs = target_kwargs
        self.arrays = []  # shm "wrapped" as np.array-objs
        self.profile = profile

        if profile:
            self.logger = mp.log_to_stderr(logging.INFO)

        for shm, shp in zip(mp_arrays, shapes):
            self.arrays.append(SharedMem.shm2ndarray(shm, shp))

    def run(self):
        while True:
            try:
                # wait till host has fetched data from shm and demands new data from this proc
                self.events['new'].wait()
                self.events['new'].clear()
                t0 = time.clock()
                result = self.target(*self.target_args, **self.target_kwargs)
                if isinstance(result, np.ndarray):
                    result = [result, ]
                for a, r in zip(self.arrays, result):
                    a[:] = r
                # signal host that task is done and data is ready in shm
                self.events['ready'].set()
                t1 = time.clock()
                if self.profile:
                    t_exec = t1 - t0
                    self.logger.info(
                        'Executing Target and writing to shm %g ms' % (
                            t_exec * 1000))
            except KeyboardInterrupt:
                pass


class BackgroundProc(SharedMem):
    """
    Data structure to manage repeated background tasks by reusing a fixed
    number of *initially* created background process with the same arguments
    at every time. (E.g. retrieving an augmented batch) Remember to call
    ``BackgroundProc.shutdown`` after use to avoid zombie process and RAM clutter.

    Parameters
    ----------

    dtypes:
      list of dtypes of the target return values
    shapes:
      list of shapes of the target return values
    n_proc: int
      number of background procs to use
    target: callable
      target function for background proc. Can even be a method of an object,
      if object data is read-only (then data will not be copied in RAM and the
      new process is lean). If several procs use random modules, new seeds
      must be created inside target because they have the same random state
      at the beginning.
    target_args:  tuple
      Proc args (constant)
    target_kwargs: dict
      Proc kwargs (constant)
    profile: Bool
      Whether to print timing results in to stdout

    Examples
    --------

    Use case to retrieve batches from a data structure ``D``:

    >>> data, label = D.getbatch(2, strided=False, flip=True, grey_augment_channels=[0])
    >>> kwargs = {'strided': False, 'flip': True, 'grey_augment_channels': [0]}
    >>> bg = BackgroundProc([np.float32, np.int16], [data.shape,label.shape], \
                            D.getbatch, n_proc=2, target_args=(2,), \
                            target_kwargs=kwargs, profile=False)
    >>> for i in range(100):
    >>>    data, label = bg.get()

    """

    def __init__(self, target, dtypes=None, shapes=None, n_proc=1,
                 target_args=(), target_kwargs={}, profile=False):
        self.dtypes = dtypes
        self.shapes = shapes
        self.target = target
        self.n_proc = n_proc
        self.i = 0  # index of next item to consume
        self.mp_arrays = []
        self.procs = []
        self.events = []
        self.profile = profile

        if (dtypes is None) or (shapes is None):
            ret = target(*target_args, **target_kwargs)
            if isinstance(ret, np.ndarray):
                ret = [ret, ]
            dtypes = [b.dtype for b in ret if b is not None]
            shapes = [b.shape for b in ret if b is not None]
            self.dtypes = dtypes
            self.shapes = shapes

        if profile:
            self.logger = mp.log_to_stderr(logging.INFO)

        # create a list of mp-arrays for each process
        for k in range(n_proc):
            a = []
            for dtype, shape in zip(dtypes, shapes):
                a.append(self.puthandle(dtype, shape))

            self.mp_arrays.append(a)
            self.events.append({'new': mp.Event(), 'ready': mp.Event()})

        # initialise the procs and give them their mp-arrays
        for shm, e in zip(self.mp_arrays, self.events):
            p = Proc(shm, shapes, e, target, target_args, target_kwargs, profile)
            p.start()
            e['new'].set()
            self.procs.append(p)

    def get(self, timeout=False):
        """
        This gets the next result from a background process and blocks
        until the corresponding proc has finished.
        """
        t0 = time.time()
        while True:  # go trough queue until a ready process is found
            k = self.i
            self.i = (self.i + 1) % self.n_proc  # advance index of next item
            if self.events[k]['ready'].is_set():
                break
            time.sleep(0.0001)
            t = time.time() - t0
            if timeout and t > timeout:
                raise TimeoutError

        result = []
        t0 = time.clock()
        self.events[k]['ready'].wait()
        self.events[k]['ready'].clear()
        t1 = time.clock()
        for shm, shp in zip(self.mp_arrays[k], self.shapes):
            # copy! Otherwise a proc will write to result
            result.append(SharedMem.shm2ndarray(shm, shp).copy())

        self.events[k]['new'].set()
        t2 = time.clock()
        if self.profile:
            t_wait = t1 - t0
            t_write = t2 - t1
            self.logger.info(
                'Waiting for subprocess %g ms, converting to numpy %g ms' % (
                    t_wait * 1000, t_write * 1000))

        return tuple(result)

    def shutdown(self):
        """**Must be called to free memory** if the background tasks are no longer needed"""
        for p in self.procs:
            p.terminate()

    def reset(self):
        """
        Should be called after an exception (e.g. by pressing ctrl+c) was raised.
        """
        for e in self.events:
            e['new'].set()

    def __del__(self):
        self.shutdown()


class SharedQ(SharedMem):
    """
    FIFO Queue to process np.ndarrays in the background
    (also pre-loading of data from disk)

    procs must accept list of ``mp.Array`` and make items ``np.ndarray``
    using ``SharedQ.shm2ndarray``, for this the shapes are required as too.
    The target requires the signature::

       >>> target(mp_arrays, shapes, *args, **kwargs)

    Whereas mp_array and shape are *automatically* added internally

    All parameters are optional:

    Parameters
    ----------
    n_proc: int
      If larger than 0, a message is printed if to few processes are running
    profile: Bool
      Whether to print timing results in terminal

    Examples
    --------
    Automatic use:

    >>> Q = SharedQ(n_proc=2)
    >>> Q.startproc(target=, shape= args=, kwargs=)
    >>> Q.startproc(target=, shape= args=, kwargs=)
    >>> for i in range(5):
    >>>     Q.startproc(target=, shape= args=, kwargs=)
    >>>     item = Q.get() # starts as many new jobs as to maintain n_proc
    >>>     dosomehtingelse(item) # processes work in background to pre-fetch data for next iteration
    """

    def __init__(self, n_proc=0, profile=False):

        self.data = deque()  # items of type [shm, shape, proc]
        self.len = 0
        self.n_proc = n_proc
        self.profile = profile

    def startproc(self, dtypes, shapes, target, target_args=(),
                  target_kwargs={}):
        """
        Starts a new process

        procs must accept list of ``mp.Array`` and make items ``np.ndarray``
        using ``SharedQ.shm2ndarray``, or this the shapes are required as
        too. The target requires the signature::

           target(mp_arrays, shapes, *args, **kwargs)

        Whereas mp_array  and shape are *automatically* added internally
        """

        data = target_kwargs.get('data')

        mp_arrays = []
        for dtype, shape in zip(dtypes, shapes):
            mp_arrays.append(self.puthandle(dtype, shape, data))

        t0 = time.clock()
        _args = (mp_arrays, shapes) + target_args
        proc = mp.Process(target=target, args=_args, kwargs=target_kwargs)
        proc.daemon = True
        proc.start()

        self.data.append([mp_arrays, shapes, proc])
        self.len += 1
        t1 = time.clock()
        if self.profile:
            t_start = t1 - t0
            self.logger.info('Start Process %g ms' % (t_start * 1000))

    def get(self):
        """
        This gets the first results in the queue and blocks until the
        corresponding proc has finished. If a n_proc value is defined this
        then new procs must be started *before* to avoid a warning message.
        """
        mp_arrays, shapes, proc = self.data.popleft()
        self.len -= 1
        missing = self.n_proc - self.len
        if missing > 0:
            self.logger.warning(
                "You should have started %i new workes before Q.get()" % missing)

        t0 = time.clock()
        proc.join()
        t1 = time.clock()
        result = []
        for shm, shp in zip(mp_arrays, shapes):
            result.append(SharedMem.shm2ndarray(shm, shp))

        t2 = time.clock()

        if self.profile:
            t_join = t1 - t0
            t_conv = t2 - t1
            self.logger.info('Join %g ms, Shared2Numpy %g ms' % (
                t_join * 1000, t_conv * 1000))

        return result


### Testing etc. ##############################################################
# Pre requisits
if __name__=="__main__":
    import gc
    import h5py


    def load():
        f = h5py.File('~/devel/data/MPI/raw_center_cube_mag1_v3.h5', 'r')
        d = f['raw'].value
        f.close()
        return d[0]


    t0 = time.time()
    D = load()
    t1 = time.time()
    lt = t1 - t0
    logger.info('REAL LOAD TIME %.2f  sec' % lt)


    def CPU():
        a = np.random.rand(1160 * 480)
        for i in range(50):
            np.sin(a)


    t0 = time.time()
    for i in range(3):
        CPU()

    t1 = time.time()
    rt = (t1 - t0) / 3
    logger.info('REAL CPU TASK TIME %.2f  sec' % rt)
    serial = 20 * rt + 20 * lt
    D = None
    gc.collect()


    def IO(mp_array, shape):
        t0 = time.time()
        d = load()
        SharedQ.shm2ndarray(mp_array, shape)[:] = d
        t = time.time() - t0
        logger.info('LOADED data in %.2f  sec' % t)




# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip

__all__ = ['AgentData', 'BatchCreatorImage', 'GridData']

import gc
import logging
import os
import sys
import time

try:
    from importlib import reload
except ImportError:
    pass  # reload() is built-in in Python 2

import h5py
import numpy as np
import tqdm

from . import transformations
from .image import make_affinities, make_nhood_targets
from .. import utils
from ..config import config

logger = logging.getLogger('elektronn2log')
inspection_logger = logging.getLogger('elektronn2log-inspection')

###############################################################################


def greyAugment(d, channels, rng):
    """
    Performs grey value (histogram) augmentations on ``d``. This is only
    applied to ``channels`` (list of channels indices), ``rng`` is a random
    number generator
    """
    if channels == []:
        return d
    else:
        k = len(channels)
        d = d.copy()  # d is still just a view, we don't want to change the original data so copy it
        alpha = 1 + (rng.rand(k) - 0.5) * 0.3 # ~ contrast
        c     = (rng.rand(k) - 0.5) * 0.3 # mediates whether values are clipped for shadows or lights
        gamma = 2.0 ** (rng.rand(k) * 2 - 1) # sample from [0.5,2] with mean 0

        d[channels] = d[channels] * alpha[:,None,None] + c[:,None,None]
        d[channels] = np.clip(d[channels], 0, 1)
        d[channels] = d[channels] ** gamma[:,None,None]
    return d


def border_treatment(data_list, ps, border_mode, ndim):
    def treatArray(data):
        if border_mode=='keep':
            return data

        sh = data.shape[1:] # (z,y,x)/(x,y)


        if border_mode=='crop':
            excess = [int((x[0] - x[1])//2) for x in zip(sh, ps)]
            if ndim == 3:
                data = data[:,
                            excess[0]:excess[0]+ps[0],
                            excess[1]:excess[1]+ps[1],
                            excess[2]:excess[2]+ps[2]]
            elif ndim==2:
                data = data[:,
                            :,
                            excess[0]:excess[0]+ps[0],
                            excess[1]:excess[1]+ps[1]]

        else:
            excess_l = [int(np.ceil(float(x[0] - x[1])/2)) for x in zip(ps, sh)]
            excess_r = [int(np.floor(float(x[0] - x[1])/2)) for x in zip(ps, sh)]
            if ndim == 3:
                pad_with = [(0,0),
                            (excess_l[0],excess_r[0]),
                            (excess_l[1],excess_r[1]),
                            (excess_l[2],excess_r[2])]
            else:
                pad_with = [(0,0),
                            (0,0),
                            (excess_l[0],excess_r[0]),
                            (excess_l[1],excess_r[1])]

            if border_mode=='mirror':
                data = np.pad(data, pad_with, mode='symmetric')

            if border_mode=='0-pad':
                data = np.pad(data, pad_with, mode='constant', constant_values=0)

        return data

    return [treatArray(d) for d in data_list]


class BatchCreatorImage(object):
    def __init__(self, input_node, target_node=None, d_path=None, l_path=None,
                 d_files=None, l_files=None, cube_prios=None, valid_cubes=None,
                 border_mode='crop', aniso_factor=2, target_vec_ix=None,
                 target_discrete_ix=None, h5stream=False):

        assert (d_path and l_path and d_files and l_files)
        if len(d_files)!=len(l_files):
            raise ValueError("d_files and l_files must be lists of same length!")
        d_path = os.path.expanduser(d_path)
        l_path = os.path.expanduser(l_path)

        self.h5stream = h5stream
        self.d_path = d_path
        self.l_path = l_path
        self.d_files = d_files
        self.l_files = l_files
        self.cube_prios = cube_prios
        self.valid_cubes = valid_cubes if valid_cubes is not None else []
        self.aniso_factor  = aniso_factor
        self.border_mode = border_mode
        self.target_vec_ix = target_vec_ix
        self.target_discrete_ix = target_discrete_ix

        # Infer geometric info from input/target shapes
        self.ndim = input_node.shape.ndim
        self.patch_size = np.array(input_node.shape.spatial_shape, dtype=np.int)
        self.strides = np.array(target_node.shape.strides, dtype=np.int)
        self.offsets = np.array(target_node.shape.offsets, dtype=np.int)
        self.target_ps = self.patch_size - self.offsets * 2
        self.t_dtype = target_node.output.dtype
        if input_node.shape.ndim==target_node.shape.ndim:
            self.mode = 'img-img'
        else:
            if target_node.shape.ndim > 1:
                raise ValueError() ###TODO would it work to map 'img-vect'?
            self.mode = 'img-scalar'

        # The following will be inferred when reading data
        self.n_labelled_pixel = 0
        self.n_f     = None # number of channels/feature in input
        self.t_n_f  = None # the shape of the returned label batch at index 1

        # Actual data fields
        self.valid_d = []
        self.valid_l = []
        self.valid_extra = []

        self.train_d = []
        self.train_l = []
        self.train_extra = []

        # Setup internal stuff
        self.rng = np.random.RandomState(np.uint32((time.time() * 0.0001 -
                                                    int(
                                                        time.time() * 0.0001)) * 4294967295))
        self.pid = os.getpid()
        self.gc_count = 1

        self._sampling_weight = None
        self._training_count = None
        self._valid_count = None
        self.n_successful_warp = 0
        self.n_failed_warp = 0

        self.load_data()
        logger.info('{}\n'.format(self.__repr__()))

    def __repr__(self):
        s = "{0:,d}-target Data Set with {1:,d} input channel(s):\n"+\
        "#train cubes: {2:,d} and #valid cubes: {3:,d}, {4:,d} labelled "+\
        "pixels."
        s = s.format(self.t_n_f, self.n_f, self._training_count,
                                  self._valid_count, self.n_labelled_pixel)
        return s

    @property
    def warp_stats(self):
        return "Warp stats: successful: %i, failed %i, quota: %.1f" %(
            self.n_successful_warp, self.n_failed_warp,
            float(self.n_successful_warp)/(self.n_failed_warp+self.n_successful_warp))

    def _reseed(self):
        """Reseeds the rng if the process ID has changed!"""
        current_pid = os.getpid()
        if current_pid!=self.pid:
            self.pid = current_pid
            self.rng.seed(np.uint32((time.time()*0.0001 -
                                     int(time.time()*0.0001))*4294967295+self.pid))
            reload(transformations) # needed for numba GIL released stuff to work
            logger.debug("Reseeding RNG in Process with PID: {}".format(self.pid))

    def _allocbatch(self, batch_size):
        images = np.zeros((batch_size, self.n_f,)+tuple(self.patch_size), dtype='float32')
        sh = self.patch_size - self.offsets * 2
        target  = np.zeros((batch_size, self.t_n_f)+tuple(sh), dtype=self.t_dtype)
        return images, target


    def getbatch(self, batch_size=1, source='train',
                 grey_augment_channels=None, warp=False, warp_args=None,
                 ignore_thresh=False, force_dense=False,
                 affinities=False, nhood_targets=False, ret_ll_mask=False):
        """
        Prepares a batch by randomly sampling, shifting and augmenting
        patches from the data

        Parameters
        ----------
        batch_size: int
            Number of examples in batch (for CNNs often just 1)
        source: str
            Data set to draw data from: 'train'/'valid'
        grey_augment_channels: list
            List of channel indices to apply grey-value augmentation to
        warp: bool or float
            Whether warping/distortion augmentations are applied to examples
            (slow --> use multiprocessing). If this is a float number,
            warping is applied to this fraction of examples e.g. 0.5 --> every
            other example.
        warp_args: dict
            Additional keyword arguments that get passed through to
            elektronn2.data.transformations.get_warped_slice()
        ignore_thresh: float
            If the fraction of negative targets in an example patch exceeds this
            threshold, this example is discarded (Negative targets are ignored
            for training [but could be used for unsupervised target propagation]).
        force_dense: bool
            If True the targets are *not* sub-sampled according to the CNN output\
            strides. Dense targets requires MFP in the CNN!
        affinities
        nhood_targets
        ret_ll_mask: bool
            If True additional information for reach batch example is returned.
            Currently implemented are two ll_mask arrays to indicate the targetling mode.
            The first dimension of those arrays is the batch_size!

        Returns
        -------
        data: np.ndarray
            [bs, ch, x, y] or [bs, ch, z, y, x] for 2D and 3D CNNS
        target: np.ndarray
            [bs, ch, x, y] or [bs, ch, z, y, x]
        ll_mask1: np.ndarray
            (optional) [bs, n_target]
        ll_mask2: np.ndarray
            (optional) [bs, n_target]
        """
        # This is especially required for multiprocessing
        if grey_augment_channels is None:
            grey_augment_channels = []
        self._reseed()
        images, target = self._allocbatch(batch_size)
        ll_masks = []
        patch_count = 0
        while patch_count < batch_size: # Loop to fill up batch with examples
            d, t, ll_mask = self._getcube(source) # get cube randomly

            try:
                if self.mode=='img-img':
                    d, t = self.warp_cut(d, t, warp, warp_args)
                else:
                    d, _ = self.warp_cut(d, None, warp, warp_args)
                self.n_successful_warp += 1

            except transformations.WarpingOOBError:
                    self.n_failed_warp += 1
                    continue

            # Check only if a ignore_thresh is set and the cube is labelled
            if (ignore_thresh is not False) and (not np.any(ll_mask[1])):
                if (t<0).mean() > ignore_thresh:
                    continue # do not use cubes which have no information

            if source == "train": # no grey augmentation for testing
                d = greyAugment(d, grey_augment_channels, self.rng)

            target[patch_count] = t
            images[patch_count] = d
            ll_masks.append(ll_mask)
            patch_count += 1

        # Patch / Nhood target stuff
        if nhood_targets is not None and nhood_targets is not False:
            if isinstance(nhood_targets, np.ndarray):
                nhood = nhood_targets
            else:
                nhood = np.array([[0, 0, 0],
                                  [1, 0, 0],
                                  [-1, 0, 0],
                                  [0, 1, 0],
                                  [0, -1, 0],
                                  [0, 0, 1],
                                  [0, 0, -1]], dtype=np.int)

            if target.shape[1]==1:
                target = make_nhood_targets(target, nhood)
            else:  # Nhood targets for first feature only, keep other targets
                tmp = make_nhood_targets(target[:, 0:1], nhood)
                target = np.concatenate([tmp, target[:, 1:]], axis=1)


        # Final modification of targets: striding and replacing nan
        if not (force_dense or np.all(self.strides==1)):
            target = self._stridedtargets(target)

        ret = [images, target]  # The "normal" batch

        # MALIS stuff
        if affinities == 'malis':
            # create aff from seg IDs (i.e. shape[1] must be 1)
            aff, seg = make_affinities(target[:, 0])  # [bs, 3, z, y, x]
            seg = seg[:, None]  # add 'class' dimension
            ret = [images, aff, seg]
        elif affinities == 'affinity':
            aff, seg = make_affinities(target)  # [bs, z, y, x, 3]
            ret = [images, aff]

        # Lazy Labels stuff
        if ret_ll_mask:  # ll_mask is now a list(bs) of tuples(2)
            ll_masks = np.atleast_3d(np.array(ll_masks, dtype=np.int16))  # (bs, 2, 5)
            ll_mask1 = ll_masks[:, 0]
            ll_mask2 = ll_masks[:, 1]
            ret += [ll_mask1, ll_mask2]

        target[np.isnan(target)] = -666  # hack because theano does not support nan


        self.gc_count += 1
        if self.gc_count%1000==0:
            gc.collect()

        return tuple(ret)

    def warp_cut(self, img, target, warp, warp_params):
        """
        (Wraps :py:meth:`elektronn2.data.transformations.get_warped_slice()`)
        
        Cuts a warped slice out of the input and target arrays.
        The same random warping transformation is each applied to both input
        and target.
        
        Warping is randomly applied with the probability defined by the ``warp``
        parameter (see below).
        
        Parameters
        ----------
        img: np.ndarray
            Input image
        target: np.ndarray
            Target image
        warp: float or bool
            False/True disable/enable warping completely.
            If ``warp`` is a float, it is used as the ratio of inputs that
            should be warped.
            E.g. 0.5 means approx. every second call to this function actually
            applies warping to the image-target pair.
        warp_params: dict
            kwargs that are passed through to
            :py:meth:`elektronn2.data.transformations.get_warped_slice()`.
            Can be empty.
        
        Returns
        -------
        d: np.ndarray
            (Warped) input image slice
        t: np.ndarray
            (Warped) target slice
        """
        if (warp is True) or (warp==1): # always warp
            do_warp=True
        elif (0 < warp < 1): # warp only a fraction of examples
            do_warp=True if (self.rng.rand()<warp) else False
        else: # never warp
            do_warp=False

        if not do_warp:
            warp_params = dict(warp_params)
            warp_params['warp_amount'] = 0

        d, t = transformations.get_warped_slice(img, self.patch_size,
                            aniso_factor=self.aniso_factor, target=target,
                            target_ps=self.target_ps, target_vec_ix=self.target_vec_ix,
                            target_discrete_ix=self.target_discrete_ix,
                                                rng=self.rng, **warp_params)

        return d, t


    def _getcube(self, source):
        """
        Draw an example cube according to sampling weight on training data,
        or randomly on valid data
        """
        if source=='train':
            p = self.rng.rand()
            i = np.flatnonzero(self._sampling_weight <= p)[-1]
            d, t, ll_mask = self.train_d[i], self.train_l[i], self.train_extra[i]
        elif source == "valid":
            if len(self.valid_d) == 0:
                logger.info("Validation Set empty. Disable testing on validation set.")
                raise ValueError("No validation set")

            i = self.rng.randint(0, len(self.valid_d))
            d = self.valid_d[i]
            t = self.valid_l[i]
            ll_mask = self.valid_extra[i]
        else:
            raise ValueError("Unknown data source")

        return d, t, ll_mask


    def _stridedtargets(self, lab):
        if self.ndim == 3:
            return lab[:, :, ::self.strides[0], ::self.strides[1], ::self.strides[2]]
        elif self.ndim == 2:
            return lab[:, :, ::self.strides[0], ::self.strides[1]]

    def load_data(self):
        """
        Parameters
        ----------

        d_path/l_path: string
          Directories to load data from
        d_files/l_files: list
          List of data/target files in <path> directory (must be in the same order!).
          Each list element is a tuple in the form
          **(<Name of h5-file>, <Key of h5-dataset>)**
        cube_prios: list
          (not normalised) list of sampling weights to draw examples from
          the respective cubes. If None the cube sizes are taken as priorities.
        valid_cubes: list
          List of indices for cubes (from the file-lists) to use as validation
          data and exclude from training, may be empty list to skip performance
          estimation on validation data.
        """
        # returns lists of cubes, ll_mask is a tuple per cube
        data, target, extras = self.read_files()

        if self.mode=='img-scalar':
            data = border_treatment(data, self.patch_size, self.border_mode,
                                    self.ndim)

        default_extra = (np.ones(self.t_n_f), np.zeros(self.t_n_f))
        extras = [default_extra if x is None else x for x in extras]

        prios = []
        # Distribute Cubes into training and valid list
        for k, (d, t, e) in enumerate(zip(data, target, extras)):
            if k in self.valid_cubes:
                self.valid_d.append(d)
                self.valid_l.append(t)
                self.valid_extra.append(e)
            else:
                self.train_d.append(d)
                self.train_l.append(t)
                self.train_extra.append(e)
                # If no priorities are given: sample proportional to cube size
                prios.append(t.size)

        if self.cube_prios is None:
            prios = np.array(prios, dtype=np.float)
        else: # If priorities are given: sample irrespective of cube size
            prios = np.array(self.cube_prios, dtype=np.float)

        # sample example i if: batch_prob[i] < p
        self._sampling_weight = np.hstack((0, np.cumsum(prios / prios.sum())))
        self._training_count = len(self.train_d)
        self._valid_count = len(self.valid_d)


    def check_files(self):
        """
        Check if file paths in the network config are available.
        """
        notfound = False
        give_neuro_data_hint = False
        fullpaths = [os.path.join(self.d_path, f) for f, _ in self.d_files] +\
                    [os.path.join(self.l_path, f) for f, _ in self.l_files]
        for p in fullpaths:
            if not os.path.exists(p):
                print('{} not found.'.format(p))
                notfound = True
                if 'neuro_data_zxy' in p:
                    give_neuro_data_hint = True
        if give_neuro_data_hint:
            print('\nIt looks like you are referencing the neuro_data_zxy dataset.\n'
                  'To install the neuro_data_xzy dataset to the default location, run:\n'
                  '  $ wget http://elektronn.org/downloads/neuro_data_zxy.zip\n'
                  '  $ unzip neuro_data_zxy.zip -d ~/neuro_data_zxy')
        if notfound:
            print('\nPlease fetch the necessary dataset and/or '
                  'change the relevant file paths in the network config.')
            sys.stdout.flush()
            sys.exit(1)

    def read_files(self):
        """
        Image files on disk are expected to be in order (ch,x,y,z) or (x,y,z)
        But image stacks are returned as (z,ch,x,y) and target as (z,x,y,)
        irrespective of the order in the file. If the image files have no
        channel this dimension is extended to a singleton dimension.
        """
        self.check_files()
        data, target, extras = [], [], []
        pbar = tqdm.tqdm(total=len(self.d_files), ncols=120, leave=False)


        for (d_f, d_key), (l_f, l_key) in zip(self.d_files, self.l_files):
            pbar.write('Loading %s and %s' % (d_f,l_f))
            if self.h5stream:
                d = h5py.File(os.path.join(self.d_path, d_f), 'r')[d_key]
                t = h5py.File(os.path.join(self.l_path, l_f), 'r')[l_key]
                assert d.compression==t.compression==None
                assert len(d.shape)==len(t.shape)==4
                assert d.dtype==np.float32
                assert t.dtype==self.t_dtype

            else:
                d = utils.h5load(os.path.join(self.d_path, d_f), d_key)
                t = utils.h5load(os.path.join(self.l_path, l_f), l_key)

            try:
                ll_mask_1 = utils.h5load(os.path.join(self.l_path, l_f), 'll_mask')
                extras.append(ll_mask_1)
            except KeyError:
                extras.append(None)

            if self.mode == 'img-scalar':
                assert t.ndim == 1, "Scalar targets must be 1d"

            if len(d.shape) == 4: # h5 dataset has no ndim
                self.n_f = d.shape[0]
            elif len(d.shape) == 3:  # We have no channels in data
                self.n_f = 1
                d = d[None]  # add (empty) 0-axis


            if len(t.shape) == 3: # If labels not empty add first axis
                t = t[None]

            if self.t_n_f is None:
                self.t_n_f = t.shape[0]
            else:
                assert self.t_n_f == t.shape[0]


            self.n_labelled_pixel += t[0].size

            # determine normalisation depending on int or float type
            if d.dtype in [np.int, np.int8, np.int16, np.int32, np.uint32,
                           np.uint, np.uint8, np.uint16, np.uint32, np.uint32]:
                m = 255
                d = np.ascontiguousarray(d, dtype=np.float32) / m

            if (np.dtype(self.t_dtype) is not np.dtype(t.dtype)) and \
                self.t_dtype not in ['float32']:
                m = t.max()
                M = np.iinfo(self.t_dtype).max
                if m  > M:
                    raise ValueError("Loading of data: targets must be cast "
                                     "to %s, but %s cannot store value %g, "
                                     "maximum allowed value: %g. You may try "
                                     "to renumber targets." %(self.t_dtype,
                                                             self.t_dtype, m, M))
            if not self.h5stream:
                d = np.ascontiguousarray(d, dtype=np.float32)
                t = np.ascontiguousarray(t, dtype=self.t_dtype)

            pbar.write('Shapes: data %s, targets %s' % (d.shape, t.shape))

            data.append(d)
            target.append(t)
            gc.collect()
            pbar.update()

        return data, target, extras

"""
3d Label kinds:
k: number of classes
n_c: number of categories (each of which has n_k classes)

normal single category multiclass classification (e.g. bg,syn,mito,ves):
    (x,y,z)   encoded_targets==True, n_targets==k
    (1,x,y,z) encoded_targets==True, n_targets==k
    (k,x,y,z) encoded_targets==False, n_targets==k
multicategory multiclass classification (e.g. affinities):
    (n_c, x,y,z)  encoded_targets==True, n_targets=k
    (n_c*k,x,y,z) encoded_targets==False, n_targets=k,
single target regression (e.g. image reconstruction of gray img):
    (x,y,z)   encoded_targets indiff, n_targets=1
    (1,x,y,z) encoded_targets indiff, n_targets=1
multi target regression (e.g. image reconstruction of rgb img):
    (k,x,y,z) n_targets=k

--> l.shape[1] = ?
"""


##############################################################################################################
class DummySkel(object):
    def __init__(self):
        self.grad = np.array([0,0,0], dtype=np.float32)
        self.last_slicing_params = None
        self.lost_track = True # don't train seqentially on this
        self.debug_traces = []
        self.debug_traces_current = []
        self.debug_grads = []
        self.debug_grads_current = []

    def _grad(self, pred_zxy):
        # L= 1/2(x**2 + y**2)
        # dL/dp = [0, x, y]
        return np.array([0,pred_zxy[1],pred_zxy[2]], dtype=np.float32)

    def get_loss_and_gradient(self, prediction_c, **kwargs):
        loss = abs(prediction_c[1:]).sum() # sum of x,y magnitude
        loss = np.array([loss, ], dtype=np.float32)

        self.debug_traces_current.append(None)
        self.debug_grads_current.append(None)

        return loss, self._grad(prediction_c)

class AgentData(BatchCreatorImage):
    """
    Load raw_cube, vec_prob_obj_cube and skelfiles + rel.offset

    """
    def __init__(self, input_node, side_target_node, path_prefix=None,
                 raw_files=None, skel_files=None, vec_files=None, valid_skels=None,
                 target_vec_ix=None, target_discrete_ix=None,
                 abs_offset=None, aniso_factor=2):

        assert (path_prefix and raw_files and skel_files and vec_files and abs_offset)
        abs_offset = np.array(abs_offset, dtype=np.int)


        path_prefix = os.path.expanduser(path_prefix)
        self.path_prefix = path_prefix
        self.raw_files = raw_files
        self.skel_files = skel_files
        self.vec_files = vec_files
        self.valid_skels = valid_skels if valid_skels is not None else []
        self.abs_offset = abs_offset
        self.aniso_factor  = aniso_factor
        self.target_vec_ix = target_vec_ix
        self.target_discrete_ix = target_discrete_ix

        # Infer geometric info from input/target shapes
        self.ndim = input_node.shape.ndim
        self.patch_size = np.array(input_node.shape.spatial_shape, dtype=np.int)
        self.strides = np.array(side_target_node.shape.strides, dtype=np.int)
        self.offsets = np.array(side_target_node.shape.offsets, dtype=np.int)
        self.target_ps = self.patch_size - self.offsets * 2
        #self.n_ch_model = input_node.shape['f']
        self.n_f       = None # number of channels/feature in input
                              # will be inferred when reading data
        self.t_n_f  = None # the shape of the returned label batch at index 1
        self.t_dtype = side_target_node.output.dtype
        # Need to infer this when reading lables and with n_target/encoded_targets

        # Setup internal stuff
        self.rng = np.random.RandomState(np.uint32((time.time()*0.0001 -
                                            int(time.time()*0.0001))*4294967295))
        self.pid = os.getpid()
        self.gc_count = 1
        self.n_failed_warp = 0
        self.n_successful_warp = 0

        self._sampling_weight = None
        self._training_count  = None
        self._valid_count     = None

        # Actual data fields
        self.train_d = []
        self.train_l = []
        self.valid_s = []
        self.train_s = []

        assert len(raw_files)==len(skel_files)==len(vec_files)==1, "not supported atm, skeletons " \
                                                                   "from different cubes are concated, " \
                                                                   "only one abs offset"
        self.load_data()
        logger.info('{}\n'.format(self.__repr__()))

    def __repr__(self):
        s = "Data Set with {0:,d} input channel(s):\n"+\
        "#train skels: {1:,d} and #valid skels: {2:,d}"
        s = s.format(self.n_f, len(self.train_s),
                     len(self.valid_s))
        return s

    def _allocbatch(self, batch_size):
        images, target = super(AgentData, self)._allocbatch(batch_size)
        skel_i = np.zeros(batch_size, dtype='int16')
        slicing_params = np.zeros([batch_size,20], dtype='float32')
        return images, target, skel_i, slicing_params

    def getbatch(self, batch_size=1, source='train', aniso=True,
                  z_shift=0, gamma=0,
                  grey_augment_channels=None,
                  r_max_scale=0.9,
                  tracing_dir_prior_c=0.5,
                  force_dense=False, flatfield_p=1e-3):

        # This is especially required for multiprocessing
        if grey_augment_channels is None:
            grey_augment_channels = []
        self._reseed()
        images, target, skel_i, slicing_params = self._allocbatch(batch_size)
        patch_count = 0
        while patch_count < batch_size: # Loop to fill up batch with examples
            if (self.rng.rand() > flatfield_p) or source!='train':
                skel, i = self.getskel(source) # get skel randomly
                position_s = skel.sample_tube_point(self.rng, r_max_scale=r_max_scale)
                local_direc_is = skel.sample_local_direction_iso(position_s, n_neighbors=6)
                tracing_direc_is = skel.sample_tracing_direction_iso(self.rng, local_direc_is, c=tracing_dir_prior_c)

                position_l = position_s[::-1]  # from lab to data frame (xyz) -> (zxy)
                position_d = position_l - self.abs_offset
                tracing_direc_il = tracing_direc_is[::-1]         # from lab to data frame (xyz) -> (zxy)

                try:
                    if config.inspection > 2: # mark position by opaque cube
                        self.train_d[0][:,
                            position_d[0] - 1:position_d[0] + 2,
                            position_d[1] - 2:position_d[1] + 3,
                            position_d[2] - 2:position_d[2] + 3] = 1.0

                    image, vec, M = transformations.get_tracing_slice(
                    self.train_d[0], self.patch_size, position_d, z_shift=z_shift,
                    aniso_factor=self.aniso_factor, sample_aniso=aniso,
                    gamma=gamma, direction_iso=tracing_direc_il,
                    target=self.train_l[0], target_ps=self.target_ps,
                    target_vec_ix=self.target_vec_ix,
                    target_discrete_ix=self.target_discrete_ix, rng=self.rng)
                    self.n_successful_warp += 1

                    trafo = transformations.Transform(M, position_l=position_l,
                                                      aniso_factor=self.aniso_factor)

                except transformations.WarpingOOBError:
                    self.n_failed_warp += 1
                    continue

                if source == "train": # no grey augmentation for testing
                    image    = greyAugment(image, grey_augment_channels, self.rng)

            else:
                vec_sh = (self.train_l[0].shape[0],)+tuple(self.target_ps)
                vec = np.ones(vec_sh, dtype=self.t_dtype) * np.nan
                img_sh = (self.train_d[0].shape[0],)+tuple(self.patch_size)
                image = np.zeros(img_sh, dtype=np.float32)
                i = len(self.train_s) - 1
                M = np.eye(4)
                position_l = np.zeros(3, np.int)
                position_d = position_l
                local_direc_is = position_l
                tracing_direc_il = position_l
                trafo = transformations.Transform(M, position_l=position_l,
                                                  aniso_factor=self.aniso_factor)

            target[patch_count] = vec
            images[patch_count] = image
            skel_i[patch_count] = i
            slicing_params[patch_count] = trafo.to_array()
            patch_count += 1

        target[np.isnan(target)] = -666 # Stupid hack because theano does not support nan
        if not (force_dense or np.all(self.strides==1)):
            target = self._stridedtargets(target)

        ret = [images, target, skel_i, slicing_params]

        if config.inspection > 2:
            debug_img = self.train_d[0][:,
                        position_d[0] - 11:position_d[0] + 12,
                        position_d[1] - 21:position_d[1] + 22,
                        position_d[2] - 21:position_d[2] + 22]
            if i == len(self.train_s) - 1:
                i = "DummySkel"

            local_direc_il = local_direc_is[::-1]
            inspection_logger.info("skeleton: %s" % i)
            inspection_logger.info("pos_d: %s, pos_l: %s, ldir_il: %s, tdir_il: %s"
                                   % (position_d.tolist(),
                                      position_l.tolist(),
                                      local_direc_il.tolist(),
                                      tracing_direc_il.tolist()))

            inspection_logger.info("params: %s" % (trafo.to_array().tolist()))

            ret = [images, target, skel_i, slicing_params, debug_img]

        self.gc_count += 1
        if self.gc_count%1000==0:
            gc.collect()

        return tuple(ret)


    def get_newslice(self, position_l, direction_il, batch_size=1, source='train',
                     aniso=True, z_shift=0, gamma=0, grey_augment_channels=None,
                     r_max_scale=0.9, tracing_dir_prior_c=0.5, force_dense=False,
                     flatfield_p=1e-3, scale=1.0, last_ch_max_interp=False):
        assert batch_size==1
        if grey_augment_channels is None:
            grey_augment_channels = []
        if source!='train':
            raise ValueError

        images, target, skel_i, slicing_params = self._allocbatch(batch_size)

        position_d = position_l - self.abs_offset
        if config.inspection > 2:  # mark position by opaque cube
            self.train_d[0][:,
            position_d[0] - 2:position_d[0] + 3,
            position_d[1] - 4:position_d[1] + 5,
            position_d[2] - 4:position_d[2] + 5] = 1.0

        image, vec, M = transformations.get_tracing_slice(
            self.train_d[0], self.patch_size, position_d,
            z_shift=z_shift, aniso_factor=self.aniso_factor, sample_aniso=aniso,
            gamma=gamma, scale_factor=scale, direction_iso=direction_il,
            target=self.train_l[0], target_ps=self.target_ps,
            target_vec_ix=self.target_vec_ix,
            target_discrete_ix=self.target_discrete_ix, rng=self.rng,
            last_ch_max_interp=last_ch_max_interp)

        trafo = transformations.Transform(M, position_l=position_l,
                                           aniso_factor=self.aniso_factor)

        image = greyAugment(image, grey_augment_channels, self.rng)

        if vec is not None: # allow for usin this with no vec data
            target[0] = vec

        images[0] = image

        target[np.isnan(target)] = -666  # Stupid hack because theano does not support nan
        if not (force_dense or np.all(self.strides==1)):
            target = self._stridedtargets(target)

        ret = [images, target, trafo]

        if config.inspection > 2:
            debug_img = self.train_d[0][:,
                        position_d[0] - 11:position_d[0] + 12,
                        position_d[1] - 21:position_d[1] + 22,
                        position_d[2] - 21:position_d[2] + 22]

            # inspection_logger.info("skeleton: 999")
            # inspection_logger.info("pos_d: %s, pos_l: %s, ldir_il: %s, tdir_il: %s"
            #                        % (position_d.tolist(),
            #                           position_l.tolist(),
            #                           direction_il.tolist(),
            #                           direction_il.tolist()))
            # inspection_logger.info("params: %s" % (trafo.to_array().tolist()))
            image_ref, _,_= transformations.get_tracing_slice(
                self.train_d[0], self.patch_size, position_d,
                z_shift=z_shift, aniso_factor=self.aniso_factor,
                sample_aniso=aniso,
                gamma=gamma, scale_factor=1.0, direction_iso=direction_il,
                target=self.train_l[0], target_ps=self.target_ps,
                target_vec_ix=self.target_vec_ix,
                target_discrete_ix=self.target_discrete_ix, rng=self.rng)
            i = np.random.randint(0, 1014)
            utils.h5save(image, '/tmp/img%i-%.2f.h5'%(i,scale))
            utils.h5save(image_ref, '/tmp/img%i-ref.h5' % i)
            #utils.h5save(debug_img, '/tmp/dbg-img-%.2f.h5' % scale)
            ret = [images, target, trafo, debug_img]

        return tuple(ret)


    def getskel(self, source):
        """
        Draw an example skeleton according to sampling weight on training data,
        or randomly on valid data
        """
        if source=='train':
            p = self.rng.rand()
            i = np.flatnonzero(self._sampling_weight <= p)[-1]
            skel = self.train_s[i]
        elif source == "valid":
            if len(self.valid_s) == 0:
                logger.info("Validation Set empty. Disable testing on validation set.")
                return []
            i = self.rng.randint(0, len(self.valid_s))
            skel = self.valid_s[i]
        else:
            raise ValueError("Unknown data source")
        return skel, i


    def load_data(self):
        """
        Parameters
        ----------

        d_path/l_path: string
          Directories to load data from
        d_files/l_files: list
          List of data/target files in <path> directory (must be in the same order!).
          Each list element is a tuple in the form
          **(<Name of h5-file>, <Key of h5-dataset>)**
        cube_prios: list
          (not normalised) list of sampling weights to draw examples from
          the respective cubes. If None the cube sizes are taken as priorities.
        valid_cubes: list
          List of indices for cubes (from the file-lists) to use as validation
          data and exclude from training, may be empty list to skip performance
          estimation on validation data.
        """
        # returns lists of cubes, ll_mask is a tuple per cube
        data, vec, skeletons = self.read_files()
        self.train_d = data
        self.train_l = vec
        prios = []
        # Distribute Cubes into training and valid list
        for skel_list, valid_skels in zip(skeletons, self.valid_skels):
            for skel_i, skel in enumerate(skel_list):
                if skel_i in valid_skels:
                    self.valid_s.append(skel)
                else:
                    self.train_s.append(skel)
                    # If no priorities are given: sample proportional to cube size
                    prios.append(len(skel.all_nodes))


        prios = np.array(prios, dtype=np.float)
        # sample example i if: batch_prob[i] < p
        self._sampling_weight = np.hstack((0, np.cumsum(prios / prios.sum())))
        self.train_s.append(DummySkel())


    def read_files(self):
        """
        Image files on disk are expected to be in order (ch,x,y,z) or (x,y,z)
        But image stacks are returned as (z,ch,x,y) and target as (z,x,y,)
        irrespective of the order in the file. If the image files have no
        channel this dimension is extended to a singleton dimension.
        """
        logger.info("Loading Data")
        data, vec, skeletons = [], [], []

        for (d_f, d_key), (l_f, l_key), s_f in zip(self.raw_files, self.vec_files, self.skel_files):
            d = utils.h5load(os.path.join(self.path_prefix, d_f), d_key)
            if d.ndim == 4:
                self.n_f = d.shape[0]
            elif len(d.shape) == 3:  # We have no channels in data
                self.n_f = 1
                d = d[None]  # add (empty) 0-axis



            # determine normalisation depending on int or float type
            if d.dtype in [np.int, np.int8, np.int16, np.int32, np.uint32,
                           np.uint, np.uint8, np.uint16, np.uint32, np.uint32]:
                m = 255
            else:
                m  = 1

            d = np.ascontiguousarray(d, dtype=np.float32) / m

            try:
                t = utils.h5load(os.path.join(self.path_prefix, l_f), l_key)
                if len(t.shape) == 3: # If labels not empty add first axis
                    t = t[None]

                assert t.ndim==4
                self.t_n_f = t.shape[0]
                if (self.t_dtype is not t.dtype) and self.t_dtype not in ['float32']:
                    m = t.max()
                    M = np.iinfo(self.t_dtype).max
                    if m  > M:
                        raise ValueError("Loading of data: targets must be cast "
                                         "to %s, but %s cannot store value %g, "
                                         "maximum allowed value: %g. You may try "
                                         "to renumber targets." %(self.t_dtype,
                                                                 self.t_dtype, m, M))

                t = np.ascontiguousarray(t, dtype=self.t_dtype)
            except:
                t = None
                self.t_n_f = 9

            gc.collect()
            data.append(d)
            vec.append(t)
            s = utils.pickleload(os.path.join(self.path_prefix, s_f))
            skeletons.append(s)
            for skel in s: # backward fixes of attributes that are not pickled
                skel.lost_track = False
                skel.position_s = None
                skel.position_l = None
                skel.direction_il = None
                skel.start_new_training = True
                skel.prev_batch = None
                skel.trafo = None
                skel.training_traces = []
                skel.current_trace = None
                skel.linked_data = self
                skel._hull_point_bg = dict()
                skel.background_processes = config.background_processes
                skel.cnn_grid = None

        return data, vec, skeletons

class GridData(AgentData):
    def __init__(self, *args, **kwargs):
        super(GridData, self).__init__(*args, **kwargs)
        #self.mean_grid = utils.pickleload("~/CNN_Training/mean.pkl")

    def getbatch(self, **get_batch_kwargs):
        self._reseed()
        try:
            skel_example, skel_index = self.getskel('train')
            skel_example.start_new_training = True
            batch = skel_example.getbatch(0, 0.0, **get_batch_kwargs)
            img, target_img, target_grid, target_node= batch
            target_grid = target_grid[None]
            #target_grid -= self.mean_grid[None,None]
            self.last_skel = skel_example #DBG

            return img, target_grid
        except transformations.WarpingOOBError:
            return self.getbatch(**get_batch_kwargs)






# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, \
    super, zip

import logging
import os
import getpass

try:
    import colorlog
    colorize = True
except ImportError:
    colorize = False
    print('Please (pip) install colorlog.')


# Formats for colorlog.LevelFormatter
log_level_formats = {'DEBUG': '%(log_color)s%(msg)s (%(module)s:%(lineno)d)',
                     'INFO': '%(log_color)s%(msg)s',
                     'WARNING': '%(log_color)sWARNING: %(msg)s (%(module)s:%(lineno)d)',
                     'ERROR': '%(log_color)sERROR: %(msg)s (%(module)s:%(lineno)d)',
                     'CRITICAL': '%(log_color)sCRITICAL: %(msg)s (%(module)s:%(lineno)d)',}

log_colors = {'DEBUG': 'blue', 'INFO': 'cyan', 'WARNING': 'bold_yellow',
              'ERROR': 'red', 'CRITICAL': 'red,bg_white'}

log_colors_inspection = {'DEBUG': 'purple', 'INFO': 'purple',
                         'WARNING': 'yellow', 'ERROR': 'red',
                         'CRITICAL': 'red,bg_white'}

# Initialize logger that can be used
# from any submodule via logging.getLogger('elektronn2log'):
user_name = getpass.getuser()
logger = logging.getLogger('elektronn2log')
# Only set up the logger if it hasn't already been initialised before:
if not len(logger.handlers) > 0:
    logger.setLevel(logging.DEBUG)

    lfile_formatter = logging.Formatter(
        '[%(asctime)s] [%(levelname)s]\t%(message)s',
        datefmt='%Y-%m-%d %H:%M:%S')
    lfile_path = os.path.abspath('/tmp/elektronn2-%s.log' % user_name)
    lfile_level = logging.DEBUG
    lfile_handler = logging.FileHandler(lfile_path)
    lfile_handler.setLevel(lfile_level)
    lfile_handler.setFormatter(lfile_formatter)
    logger.addHandler(lfile_handler)

    if colorize:
        lstream_handler = colorlog.StreamHandler()
        lstream_handler.setFormatter(
            colorlog.LevelFormatter(fmt=log_level_formats,
                                    log_colors=log_colors))
    else:
        lstream_handler = logging.StreamHandler()  # log to stderr
    # set this to logging.DEBUG to enable output for logger.debug() calls
    lstream_level = logging.INFO
    lstream_handler.setLevel(lstream_level)
    logger.addHandler(lstream_handler)

    logger.propagate = False

    if False:  # Test log levels:
        logger.critical('== critical')
        logger.error('== error')
        logger.warning('== warn')
        logger.info('== info')
        logger.debug('== debug')

from .config import config  # import default config


inspection_logger = logging.getLogger('elektronn2log-inspection')
# Only set up the logger if it hasn't already been initialised before:
if not len(inspection_logger.handlers) > 0:
    inspection_logger.setLevel(logging.DEBUG)
    lfile_formatter = logging.Formatter('%(message)s')
    lfile_path = os.path.abspath(
        # os.path.expanduser('~/elektronn2-inspection.log')
        '/tmp/elektronn2-inspection.log'
    )
    lfile_level = logging.DEBUG
    lfile_handler = logging.FileHandler(lfile_path)
    lfile_handler.setLevel(lfile_level)
    lfile_handler.setFormatter(lfile_formatter)
    inspection_logger.addHandler(lfile_handler)

    if config.inspection:
        if colorize:
            lstream_handler = colorlog.StreamHandler()
            lstream_handler.setFormatter(
                colorlog.LevelFormatter(fmt=log_level_formats,
                                        log_colors=log_colors_inspection))
        else:
            lstream_handler = logging.StreamHandler()  # log to stderr
        # set this to logging.DEBUG to enable output for inspection_logger.debug() calls
        lstream_level = logging.INFO
        lstream_handler.setLevel(lstream_level)
        inspection_logger.addHandler(
            lstream_handler)  # comment out to supress printing

    inspection_logger.propagate = False



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip

from numpy import ndarray
import logging
from collections import OrderedDict
from functools import reduce
import theano.tensor as T

from . import node_basic

logger = logging.getLogger('elektronn2log')

__all__ = ["GraphManager"]

class NodePointer(object):
    def __init__(self, target_id):
        self.target_id = target_id

    def __repr__(self):
        return "<NodePointer> %s"%self.target_id

class ParamPointer(object):
    def __init__(self, target_id, param_name):
        self.target_id = target_id
        self.param_name = param_name

    def __repr__(self):
        return "<ParamPointer> %s of node %s" %(self.param_name, self.target_id)

class SplitDescriptor(object):
    def __init__(self, node_id, func, args, kwargs):
        self.node_id = node_id
        self.args    = args
        self.kwargs  = kwargs
        self.func    = func

    def restore(self, gm):
        node = gm.nodes[self.node_id]
        return self.func(node, *self.args, **self.kwargs)



class NodeDescriptor(object):
    @staticmethod
    def _is_node_iterable(arg):
        if hasattr(arg, '__iter__') and not isinstance(arg, T.Variable):
            return reduce(lambda x,y: x*isinstance(y, node_basic.Node), arg, True)
        else:
            return False

    @staticmethod
    def _is_node_pointer_iterable(arg):
        if hasattr(arg, '__iter__'):
            return reduce(lambda x,y: x*isinstance(y, NodePointer), arg, True)
        else:
            return False

    def __init__(self, args, kwargs, cls, gm):
        new_args = []
        for arg in args:
            if isinstance(arg, node_basic.Node): # Replace node instances by id
                arg = NodePointer(arg.name)
            elif self._is_node_iterable(arg):
                new_arg = []
                for a in arg:
                   new_arg.append(NodePointer(a.name))
                arg = new_arg

            elif isinstance(arg, T.Variable): # For T.Variable search their origin
                found = False
                for name, node in gm.nodes.items(): # Search in all nodes
                    for p_name,p in node.params.items():
                        if p is arg:
                            arg = ParamPointer(name, p_name)
                            found = True
                            break
                if not found:
                    if cls==node_basic.FromTensor:
                        logger.debug("Could not find TensorVariable positional "
                        "arg to %s node. Saving/Loading might be broken" %cls)
                    else:
                         logger.warning("Could not find TensorVariable positional "
                    "arg to %s node. Saving/Loading might be broken" %cls)

                    arg = "TensorVariable not found"

            new_args.append(arg)

        new_kwargs = dict()
        for k,kwarg in kwargs.items():
            if isinstance(kwarg, node_basic.Node): # Replace node instances by id
                kwarg = NodePointer(kwarg.name)
            elif isinstance(kwarg, T.Variable): # For T.Variable search their origin
                found = False
                for name,node in gm.nodes.items():
                    for p_name,p in node.params.items():
                        if p is kwarg:
                            kwarg = ParamPointer(name, p_name)
                            found = True
                            break
                if not found:
                    logger.warning("Could not find shared param!")

            elif isinstance(kwarg, ndarray):
                continue

            new_kwargs[k] = kwarg

        self.args = new_args
        self.kwargs = new_kwargs
        self.cls = cls


    def restore(self, param_values, gm, override_mfp_to_active=False,
                make_weights_constant=False, unlock_weights=False,
                replace_bn=False):
        if self.cls==node_basic.FromTensor:
            return None

        args = self.args
        kwargs = self.kwargs

        new_args = []
        for arg in args:
            if isinstance(arg, NodePointer): # Replace nodepointers by nodes of graph_manager
                arg = gm.nodes[arg.target_id]
            elif self._is_node_pointer_iterable(arg):
                new_arg = []
                for a in arg:
                    new_arg.append(gm.nodes[a.target_id])
                arg = new_arg
            elif isinstance(arg, ParamPointer): # lookup parameter
                arg = gm.nodes[arg.target_id].params[arg.param_name]

            new_args.append(arg)

        new_kwargs = dict()
        for k,kwarg in kwargs.items():
            if isinstance(kwarg, NodePointer): # Replace nodepointers by nodes of graph_manager
                kwarg = gm.nodes[kwarg.target_id]
            elif isinstance(kwarg,ParamPointer): # For T.Variable search their origin
                 kwarg = gm.nodes[kwarg.target_id].params[kwarg.param_name]

            new_kwargs[k] = kwarg

        if make_weights_constant: # This relies on the rule that the keyword name for a
        # param of the __init__ of a node is equal to the name that param has in
        # node.params dict (param_values is create from that)!
            for name, param in param_values.items():
                if name in ['w', 'b', 'gamma', 'mean', 'std']:
                    new_kwargs[name] = [param, 'const']

        elif unlock_weights:
            if make_weights_constant:
                raise ValueError("make_weights_constant and unlock_weights"
                                 "cannot both be True")
            for name, param in param_values.items():
                if name in ['w', 'b', 'gamma', 'mean', 'std']:
                    new_kwargs[name] = param


        if replace_bn:
            if new_kwargs.get('batch_normalisation', None):
                if replace_bn=='remove':
                    new_kwargs['batch_normalisation'] = False
                elif replace_bn=='predict':
                    new_kwargs['batch_normalisation'] = 'predict'
                    #new_kwargs['mean'] = param_values['mean'] # settind is done already below
                    #new_kwargs['std']  = param_values['std']
                elif replace_bn=='const':
                    new_kwargs['batch_normalisation'] = 'predict'
                    new_kwargs['mean'] = [param_values['mean'], 'const']
                    new_kwargs['std']  = [param_values['std'], 'const']
                else:
                    raise ValueError("replace_bn can be 'remove', 'predict' or 'const'.")

        if override_mfp_to_active:
            if self.cls.__name__ == 'Conv':
                new_kwargs['mfp'] = True

        node = self.cls(*new_args, **new_kwargs)
        node.set_param_values(param_values, skip_const=True)
        return node


class GraphManager(object):
    @staticmethod
    def get_lock_names(node_descriptors, count):
        names = []
        for name,descr in node_descriptors.items():
            if (not isinstance(descr, SplitDescriptor)) and \
            (isinstance(descr[0], NodeDescriptor)):
                if len(descr[1]): # param values
                    names.append(name)
                    if len(names)>=count:
                        break
        return names


    def __init__(self, name=""):
        self.name = name
        self.nodes = OrderedDict()
        self.node_descriptors = OrderedDict()

    def __repr__(self):
        return repr(list(self.nodes.keys()))

    def __getitem__(self, slice):
        if isinstance(slice, str):
            return self.nodes[slice]
        else:
            return list(self.nodes.values())[slice]

    def reset(self):
        self.nodes = OrderedDict()
        self.node_descriptors= OrderedDict()
        logger.debug("Cleared all nodes from the graph_manager. IDs starting from 0 again")

    def register_node(self, node, name, args, kwargs):
        self.node_descriptors[name] = NodeDescriptor(args, kwargs, node.__class__, self)
        self.nodes[name] = node # Add node only after descriptor has been created:
        # descriptor may check all existing nodes, but this not has not yet finished
        # initialisation, and so lookups might fail

    def register_split(self, node, func, name, args, kwargs):
        name = node_basic.choose_name(name, self.node_descriptors.keys())
        self.node_descriptors[name] = SplitDescriptor(node.name, func, args, kwargs)
        return name

    def serialise(self):
        descriptors = OrderedDict()
        for name,descr in self.node_descriptors.items():
            if isinstance(descr, NodeDescriptor):
                param_values = self.nodes[name].get_param_values()
                descriptor = [descr, param_values]
            else:
                descriptor = descr

            descriptors[name] = descriptor

        return descriptors


    def restore(self, descriptors, override_mfp_to_active=False,
                make_weights_constant=False, replace_bn=False,
                lock_trainable_params=False, unlock_all_params=False):

        self.reset() # This clears this gm and when the nodes are restored
        # they are registered to the current active model in the model_manager
        # so make sure that the current active model is this one!
        names_to_lock = []
        if lock_trainable_params:
            assert not unlock_all_params
            if make_weights_constant:
                logger.info("'lock_trainable_params' has no effect because"
                            "'make_weights_constant' locks all anyway.")
            if isinstance(lock_trainable_params, int):
                names_to_lock = self.get_lock_names(descriptors, lock_trainable_params)
            elif isinstance(lock_trainable_params, (list, tuple)):
                names_to_lock = list(lock_trainable_params)

            logger.info("locking parameters: %s" %(names_to_lock,))
            names_to_lock = set(names_to_lock)

        for name, descr in descriptors.items():
            # logger.debug("Processing descriptor %s:\n %s\n" % (name, rec))
            if isinstance(descr, SplitDescriptor):
                descr.restore(self)
            elif isinstance(descr[0], NodeDescriptor):
                param_values = descr[1]
                descr = descr[0]
                const = make_weights_constant or name in names_to_lock
                descr.restore(param_values, self,
                              override_mfp_to_active,
                              const, unlock_all_params, replace_bn)
            else:
                raise ValueError("Unknown descriptor: %s." %(descr))


    @property
    def sources(self):
        ret = []
        for node in self.nodes.values():
            if node.is_source:
                ret.append(node)

        return ret

    @property
    def sinks(self):
        ret = []
        for node in self.nodes.values():
            if len(node.children)==0:
                ret.append(node)

        return ret

    @property
    def node_count(self):
        return len(self.nodes)

    def plot(self):
        raise NotImplementedError()


    def function_code(self):
        raise NotImplementedError()
